from pathlib import Path

import pandas as pd
import webviz_core_components as wcc
from dash import Dash, Input, Output, html
from dash.exceptions import PreventUpdate
from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.common_cache import CACHE
from webviz_config.webviz_store import webvizstore

import webviz_subsurface._utils.parameter_response as parresp
from webviz_subsurface._models import (
    EnsembleSetModel,
    caching_ensemble_set_model_factory,
)


class ParameterParallelCoordinates(WebvizPluginABC):
    """Visualizes parameters used in FMU ensembles side-by-side. Also supports response coloring.

Useful to investigate:
* Initial parameter distributions, and convergence of parameters over multiple iterations.
* Trends in relations between parameters and responses.

!> At least two parameters have to be selected to make the plot work.

---
**Three main options for input data: Aggregated, file per realization and read from UNSMRY.**

**Using aggregated data**
* **`parameter_csv`:** Aggregated csvfile for input parameters with `REAL` and `ENSEMBLE` columns \
(absolute path or relative to config file).
* **`response_csv`:** Aggregated csvfile for response parameters with `REAL` and `ENSEMBLE` \
columns (absolute path or relative to config file).


**Using a response file per realization**
* **`ensembles`:** Which ensembles in `shared_settings` to visualize.
* **`response_file`:** Local (per realization) csv file for response parameters (Cannot be \
                    combined with `response_csv` and `parameter_csv`).
* Parameter values are extracted automatically from the `parameters.txt` files in the individual
realizations of your defined `ensembles`, using the `fmu-ensemble` library.

**Using simulation time series data directly from `UNSMRY` files as responses**
* **`ensembles`:** Which ensembles in `shared_settings` to visualize. The lack of `response_file` \
                implies that the input data should be time series data from simulation `.UNSMRY` \
                files, read using `fmu-ensemble`.
* **`column_keys`:** (Optional) slist of simulation vectors to include as responses when reading \
                from UNSMRY-files in the defined ensembles (default is all vectors). * can be \
                used as wild card.
* **`sampling`:** (Optional) sampling frequency when reading simulation data directly from \
               `.UNSMRY`-files (default is monthly).
* Parameter values are extracted automatically from the `parameters.txt` files in the individual
realizations of your defined `ensembles`, using the `fmu-ensemble` library.

?> The `UNSMRY` input method implies that the "DATE" vector will be used as a filter \
   of type `single` (as defined below under `response_filters`).

**Using the plugin without responses**
It is possible to use the plugin with only parameter data, in that case set the option \
`no_responses` to True, and give either `ensembles` or `parameter_csv` as input as described \
above. Response coloring and filtering will then not be available.

**Common settings for responses**
All of these are optional, some have defaults seen in the code snippet below.

* **`response_filters`:** Optional dictionary of responses (columns in csv file or simulation \
                       vectors) that can be used as row filtering before aggregation. \
                       Valid options:
    * `single`: Dropdown with single selection.
    * `multi`: Dropdown with multiple selection.
    * `range`: Slider with range selection.
* **`response_ignore`:** List of response (columns in csv or simulation vectors) to ignore \
                      (cannot use with response_include).
* **`response_include`:** List of response (columns in csv or simulation vectors) to include \
                       (cannot use with response_ignore).
* **`aggregation`:** How to aggregate responses per realization. Either `sum` or `mean`.

Parameter values are extracted automatically from the `parameters.txt` files in the individual
realizations of your defined `ensembles`, using the `fmu-ensemble` library.

---

?> Non-numerical (string-based) input parameters and responses are removed.

?> The responses will be aggregated per realization; meaning that if your filters do not reduce \
the response to a single value per realization in your data, the values will be aggregated \
accoording to your defined `aggregation`. If e.g. the response is a form of volume, \
and the filters are regions (or other subdivisions of the total volume), then `sum` would \
be a natural aggregation. If on the other hand the response is the pressures in the \
same volume, aggregation as `mean` over the subdivisions of the same volume \
would make more sense (though the pressures in this case would not be volume weighted means, \
and the aggregation would therefore likely be imprecise).

!> It is **strongly recommended** to keep the data frequency to a regular frequency (like \
`monthly` or `yearly`). This applies to both csv input and when reading from `UNSMRY` \
(controlled by the `sampling` key). This is because the statistics are calculated per DATE over \
all realizations in an ensemble, and the available dates should therefore not differ between \
individual realizations of an ensemble.

**Using aggregated data**

The `parameter_csv` file must have columns `REAL`, `ENSEMBLE` and the parameter columns.

The `response_csv` file must have columns `REAL`, `ENSEMBLE` and the response columns \
(and the columns to use as `response_filters`, if that option is used).


**Using a response file per realization**

Parameters are extracted automatically from the `parameters.txt` files in the individual
realizations, using the `fmu-ensemble` library.

The `response_file` must have the response columns (and the columns to use as `response_filters`, \
if that option is used).


**Using simulation time series data directly from `UNSMRY` files as responses**

Parameters are extracted automatically from the `parameters.txt` files in the individual
realizations, using the `fmu-ensemble` library.

Responses are extracted automatically from the `UNSMRY` files in the individual realizations,
using the `fmu-ensemble` library.

!> The `UNSMRY` files are auto-detected by `fmu-ensemble` in the `eclipse/model` folder of the \
individual realizations. You should therefore not have more than one `UNSMRY` file in this \
folder, to avoid risk of not extracting the right data.

"""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        app: Dash,
        webviz_settings: WebvizSettings,
        ensembles: list = None,
        parameter_csv: Path = None,
        response_csv: Path = None,
        response_file: str = None,
        response_filters: dict = None,
        response_ignore: list = None,
        response_include: list = None,
        parameter_ignore: list = None,
        column_keys: list = None,
        sampling: str = "monthly",
        aggregation: str = "sum",
        no_responses=False,
    ):

        super().__init__()

        self.parameter_csv = parameter_csv if parameter_csv else None
        self.response_csv = response_csv if response_csv else None
        self.response_file = response_file if response_file else None
        self.response_filters = response_filters if response_filters else {}
        self.column_keys = column_keys
        self.time_index = sampling
        self.aggregation = aggregation
        self.no_responses = no_responses
        self.response_columns = []

        if response_ignore and response_include:
            raise ValueError(
                'Incorrent argument. Either provide "response_include", '
                '"response_ignore" or neither'
            )
        if parameter_csv:
            if ensembles or response_file:
                raise ValueError(
                    'Incorrect arguments. Either provide "parameter_csv" or '
                    '"ensembles and/or response_file".'
                )
            if not self.no_responses:
                if self.response_csv:
                    self.responsedf = read_csv(self.response_csv)
                else:
                    raise ValueError("Incorrect arguments. Missing response_csv.")
            self.parameterdf = read_csv(self.parameter_csv)

        elif ensembles:
            if self.response_csv:
                raise ValueError(
                    'Incorrect arguments. Either provide "response_csv" or '
                    '"ensembles and/or response_file".'
                )
            self.emodel: EnsembleSetModel = (
                caching_ensemble_set_model_factory.get_or_create_model(
                    ensemble_paths={
                        ens: webviz_settings.shared_settings["scratch_ensembles"][ens]
                        for ens in ensembles
                    },
                    time_index=self.time_index,
                    column_keys=self.column_keys,
                )
            )
            self.parameterdf = self.emodel.load_parameters()
            if not self.no_responses:
                if self.response_file:
                    self.responsedf = self.emodel.load_csv(csv_file=response_file)
                else:
                    self.responsedf = self.emodel.get_or_load_smry_cached()
                    self.response_filters["DATE"] = "single"
        else:
            raise ValueError(
                "Incorrect arguments."
                'You have to define at least "ensembles" or "parameter_csv".'
            )

        if not self.no_responses:
            parresp.check_runs(parameterdf=self.parameterdf, responsedf=self.responsedf)
            parresp.check_response_filters(
                responsedf=self.responsedf, response_filters=self.response_filters
            )
            # only select numerical responses
            self.response_columns = parresp.filter_numerical_columns(
                df=self.responsedf,
                column_ignore=response_ignore,
                column_include=response_include,
                filter_columns=self.response_filters.keys(),
            )

        # Only select numerical parameters
        self.parameter_columns = parresp.filter_numerical_columns(
            df=self.parameterdf, column_ignore=parameter_ignore
        )

        self.theme = webviz_settings.theme
        self.set_callbacks(app)

    @property
    def ensembles(self):
        """Returns list of ensembles"""
        return list(self.parameterdf["ENSEMBLE"].unique())

    @property
    def ens_colormap(self):
        """Returns a discrete colormap with one color per ensemble"""
        colors = self.theme.plotly_theme["layout"]["colorway"]
        colormap = []
        for i in range(0, len(self.ensembles)):
            colormap.append([i / len(self.ensembles), colors[i]])
            colormap.append([(i + 1) / len(self.ensembles), colors[i]])

        return colormap

    @property
    def response_layout(self):
        """Layout to display selectors for response filters"""

        if self.no_responses:
            return []
        children = [
            wcc.Dropdown(
                label="Response",
                id=self.uuid("responses"),
                options=[{"label": ens, "value": ens} for ens in self.response_columns],
                clearable=False,
                value=self.response_columns[0],
                style={"marginBottom": "20px"},
            ),
        ]

        if self.response_filters is not None:
            for col_name, col_type in self.response_filters.items():
                values = list(self.responsedf[col_name].unique())
                if col_type == "multi":
                    children.append(
                        wcc.SelectWithLabel(
                            label=col_name,
                            id=self.uuid(f"filter-{col_name}"),
                            options=[{"label": val, "value": val} for val in values],
                            value=values,
                            multi=True,
                            size=min(20, len(values)),
                        )
                    )
                elif col_type == "single":
                    children.append(
                        wcc.Dropdown(
                            label=col_name,
                            id=self.uuid(f"filter-{col_name}"),
                            options=[{"label": val, "value": val} for val in values],
                            value=values[0],
                            multi=False,
                            clearable=False,
                        )
                    )

        return [
            html.Div(
                id=self.uuid("view_response"),
                style={"display": "none"},
                children=children,
            ),
        ]

    @property
    def control_layout(self):
        """Layout to select ensembles and parameters"""
        mode_select = (
            []
            if self.no_responses
            else [
                wcc.Selectors(
                    label="Mode",
                    children=[
                        wcc.RadioItems(
                            id=self.uuid("mode"),
                            options=[
                                {"label": "Ensemble", "value": "ensemble"},
                                {"label": "Response", "value": "response"},
                            ],
                            value="ensemble",
                        ),
                    ],
                )
            ]
        )

        return mode_select + [
            wcc.Selectors(
                label="Ensembles",
                children=wcc.SelectWithLabel(
                    id=self.uuid("ensembles"),
                    options=[{"label": ens, "value": ens} for ens in self.ensembles],
                    multi=True,
                    value=self.ensembles,
                    size=min(len(self.ensembles), 10),
                ),
            ),
            wcc.Selectors(
                label="Parameter filter",
                children=[
                    wcc.RadioItems(
                        id=self.uuid("exclude_include"),
                        options=[
                            {"label": "Exclude", "value": "exc"},
                            {"label": "Include", "value": "inc"},
                        ],
                        value="exc",
                    ),
                    wcc.SelectWithLabel(
                        label="Parameters",
                        id=self.uuid("parameter-list"),
                        options=[
                            {"label": ens, "value": ens}
                            for ens in self.parameter_columns
                        ],
                        multi=True,
                        size=min(len(self.parameter_columns), 15),
                        value=[],
                    ),
                ],
            ),
        ]

    @property
    def layout(self):
        """Main layout"""
        return wcc.FlexBox(
            id=self.uuid("layout"),
            style={"height": "90vh"},
            children=[
                wcc.Frame(
                    style={"flex": 1, "height": "90vh"},
                    children=(self.control_layout + self.response_layout),
                ),
                wcc.Frame(
                    style={"flex": 4, "height": "90vh", "marginRight": "2vw"},
                    color="white",
                    highlight=False,
                    children=wcc.Graph(
                        id=self.uuid("parcoords"),
                    ),
                ),
            ],
        )

    @property
    def parcoord_inputs(self):
        inputs = [
            Input(self.uuid("ensembles"), "value"),
            Input(self.uuid("exclude_include"), "value"),
            Input(self.uuid("parameter-list"), "value"),
        ]
        if not self.no_responses:
            inputs.extend(
                [
                    Input(self.uuid("mode"), "value"),
                    Input(self.uuid("responses"), "value"),
                ]
            )
            if self.response_filters is not None:
                inputs.extend(
                    [
                        Input(self.uuid(f"filter-{col}"), "value")
                        for col in self.response_filters
                    ]
                )
        return inputs

    @staticmethod
    def set_grid_layout(columns):
        return {
            "display": "grid",
            "alignContent": "space-around",
            "justifyContent": "space-between",
            "gridTemplateColumns": f"{columns}",
        }

    def set_callbacks(self, app):
        @app.callback(
            Output(self.uuid("parcoords"), "figure"),
            self.parcoord_inputs,
        )
        def _update_parcoord(ens, exc_inc, parameter_list, *opt_args):
            """Updates parallel coordinates plot
            Filter dataframe for chosen ensembles and parameters
            Call render_parcoord to render new figure
            """
            # Ensure selected ensembles is a list
            ens = ens if isinstance(ens, list) else [ens]
            # Ensure selected parameters is a list
            parameter_list = (
                parameter_list if isinstance(parameter_list, list) else [parameter_list]
            )
            special_columns = ["ENSEMBLE", "REAL"]
            if exc_inc == "exc":
                parameterdf = self.parameterdf.drop(parameter_list, axis=1)
            elif exc_inc == "inc":
                parameterdf = self.parameterdf[special_columns + parameter_list]
            params = [
                param
                for param in parameterdf.columns
                if param not in special_columns and param in self.parameter_columns
            ]

            mode = opt_args[0] if opt_args else "ensemble"
            # Need a default response
            response = ""

            if mode == "response":
                if len(ens) != 1:
                    # Need to wait for update of ensemble selector to multi=False
                    raise PreventUpdate
                df = parameterdf.loc[self.parameterdf["ENSEMBLE"] == ens[0]]
                response = opt_args[1]
                response_filter_values = opt_args[2:] if len(opt_args) > 2 else {}
                filteroptions = parresp.make_response_filters(
                    response_filters=self.response_filters,
                    response_filter_values=response_filter_values,
                )
                responsedf = parresp.filter_and_sum_responses(
                    self.responsedf,
                    ens[0],
                    response,
                    filteroptions=filteroptions,
                    aggregation=self.aggregation,
                )

                # Renaming to make it clear in plot.
                responsedf.rename(
                    columns={response: f"Response: {response}"}, inplace=True
                )
                df = pd.merge(responsedf, df, on=["REAL"]).drop(columns=special_columns)
                df[self.uuid("COLOR")] = df.apply(
                    lambda row: self.ensembles.index(ens[0]), axis=1
                )
            else:
                # Filter on ensembles (ens) and active parameters (params),
                # adding the COLOR column to the columns to keep
                df = self.parameterdf[self.parameterdf["ENSEMBLE"].isin(ens)][
                    params + ["ENSEMBLE"]
                ]
                df[self.uuid("COLOR")] = df.apply(
                    lambda row: self.ensembles.index(row["ENSEMBLE"]), axis=1
                )
            return render_parcoord(
                df,
                self.theme,
                self.ens_colormap,
                self.uuid("COLOR"),
                self.ensembles,
                mode,
                params,
                response,
            )

        @app.callback(
            [
                Output(self.uuid("ensembles"), "multi"),
                Output(self.uuid("ensembles"), "value"),
                Output(self.uuid("view_response"), "style"),
            ],
            [Input(self.uuid("mode"), "value")],
        )
        def _update_mode(mode: str):
            if mode == "ensemble":
                return True, self.ensembles, {"display": "none"}
            if mode == "response":
                return False, self.ensembles[0], {"display": "block"}
            # The error should never occur
            raise ValueError("ensemble and response are the only valid modes.")

    def add_webvizstore(self):
        functions = []
        if self.parameter_csv:
            functions.append(
                (
                    read_csv,
                    [
                        {
                            "csv_file": self.parameter_csv,
                        }
                    ],
                )
            )
            if self.response_csv:
                functions.append(
                    (
                        read_csv,
                        [
                            {
                                "csv_file": self.response_csv,
                            }
                        ],
                    )
                )
        else:
            functions.extend(self.emodel.webvizstore)

        return functions


def render_parcoord(plot_df, theme, colormap, color_col, ens, mode, params, response):
    """Renders parallel coordinates plot"""
    colormap = (
        colormap if mode == "ensemble" else theme.plotly_theme["layout"]["colorway"]
    )
    if response:
        response = f"Response: {response}"
        params = [response] + params
    # Create parcoords dimensions (one per parameter)
    dimensions = [{"label": param, "values": plot_df[param]} for param in params]
    data = [
        {
            "line": {
                "color": plot_df[color_col].values.tolist(),
                "colorscale": colormap,
                "cmin": -0.5,
                "cmax": len(ens) - 0.5,
                "showscale": True,
                "colorbar": {
                    "tickvals": list(range(0, len(ens))),
                    "ticktext": ens,
                    "title": "Ensemble",
                    "xanchor": "right",
                    "x": -0.02,
                    "len": 0.2 * len(ens),
                },
            },
            "dimensions": dimensions,
            "labelangle": 60,
            "labelside": "bottom",
            "type": "parcoords",
        }
        if mode == "ensemble"
        else {
            "type": "parcoords",
            "line": {
                "color": plot_df[response],
                "colorscale": colormap,
                "showscale": True,
                "colorbar": {
                    "title": {"text": response},
                    "xanchor": "right",
                    "x": -0.02,
                },
            },
            "dimensions": dimensions,
            "labelangle": 60,
            "labelside": "bottom",
        }
        if mode == "response"
        else {}
    ]

    # Ensure sufficient spacing between each dimension and margin for labels
    width = len(dimensions) * 80 + 250
    layout = {"width": width, "height": 1200, "margin": {"b": 740, "t": 30}}
    return {"data": data, "layout": theme.create_themed_layout(layout)}


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def read_csv(csv_file) -> pd.DataFrame:
    return pd.read_csv(csv_file, index_col=False)



from typing import Dict, List

import pandas as pd

from webviz_subsurface._utils.ensemble_summary_provider_set import (
    EnsembleSummaryProviderSet,
)
from webviz_subsurface.plugins._simulation_time_series._views._subplot_view._types import (
    DeltaEnsemble,
)

# pylint: disable = line-too-long
from webviz_subsurface.plugins._simulation_time_series._views._subplot_view._utils.derived_ensemble_vectors_accessor_utils import (
    create_derived_vectors_accessor_dict,
)

# pylint: disable = line-too-long
from webviz_subsurface.plugins._simulation_time_series._views._subplot_view._utils.derived_vectors_accessor.derived_delta_ensemble_vectors_accessor_impl import (
    DerivedDeltaEnsembleVectorsAccessorImpl,
)

# pylint: disable = line-too-long
from webviz_subsurface.plugins._simulation_time_series._views._subplot_view._utils.derived_vectors_accessor.derived_ensemble_vectors_accessor_impl import (
    DerivedEnsembleVectorsAccessorImpl,
)
from webviz_subsurface.plugins._simulation_time_series._views._subplot_view._utils.derived_vectors_accessor.derived_vectors_accessor import (
    DerivedVectorsAccessor,
)

from ..mocks.derived_vectors_accessor_ensemble_summary_provider_mock import (
    EnsembleSummaryProviderMock,
)


def test_create_derived_vectors_accessor_dict() -> None:
    ensembles = ["ensA", "ensB", "(ensA)-(ensB)", "(ensC)-(ensA)"]
    vectors = ["vector_1", "vector_2"]
    provider_set = EnsembleSummaryProviderSet(
        {
            "ensA": EnsembleSummaryProviderMock(pd.DataFrame()),
            "ensB": EnsembleSummaryProviderMock(pd.DataFrame()),
            "ensC": EnsembleSummaryProviderMock(pd.DataFrame()),
        }
    )
    delta_ensembles: List[DeltaEnsemble] = [
        DeltaEnsemble(ensemble_a="ensA", ensemble_b="ensB"),
        DeltaEnsemble(ensemble_a="ensC", ensemble_b="ensA"),
    ]

    created_result: Dict[
        str, DerivedVectorsAccessor
    ] = create_derived_vectors_accessor_dict(
        ensembles=ensembles,
        vectors=vectors,
        provider_set=provider_set,
        expressions=[],
        delta_ensembles=delta_ensembles,
        resampling_frequency=None,
        relative_date=None,
    )

    assert len(created_result) == len(ensembles)

    assert created_result.get("ensA", None) is not None
    assert isinstance(created_result["ensA"], DerivedEnsembleVectorsAccessorImpl)

    assert created_result.get("ensB", None) is not None
    assert isinstance(created_result["ensB"], DerivedEnsembleVectorsAccessorImpl)

    assert created_result.get("(ensA)-(ensB)", None) is not None
    assert isinstance(
        created_result["(ensA)-(ensB)"], DerivedDeltaEnsembleVectorsAccessorImpl
    )

    assert created_result.get("(ensC)-(ensA)", None) is not None
    assert isinstance(
        created_result["(ensC)-(ensA)"], DerivedDeltaEnsembleVectorsAccessorImpl
    )



import datetime
from typing import Any, Dict, List, Optional, Set, Union

import plotly.graph_objects as go
from dash import ALL, Input, Output, State, callback, callback_context
from dash.development.base_component import Component
from webviz_config import WebvizConfigTheme
from webviz_config.utils import StrEnum, callback_typecheck
from webviz_config.webviz_plugin_subclasses import ViewABC

from ..._types import ChartType
from ..._utils import EnsembleWellAnalysisData
from ._settings import (
    WellOverviewChartType,
    WellOverviewFilters,
    WellOverviewLayoutOptions,
    WellOverviewSelections,
)
from ._utils import WellOverviewFigure, format_well_overview_figure
from ._view_element import WellOverviewViewElement


class WellOverviewView(ViewABC):
    class Ids(StrEnum):
        CHART_TYPE = "chart-type"
        SELECTIONS = "selections"
        LAYOUT_OPTIONS = "layout-options"
        FILTERS = "filters"
        CURRENT_FIGURE = "current-figure"
        VIEW_ELEMENT = "view-element"

    def __init__(
        self,
        data_models: Dict[str, EnsembleWellAnalysisData],
        theme: WebvizConfigTheme,
    ) -> None:
        super().__init__("Well overview")

        self._data_models = data_models
        self._theme = theme

        ensembles = list(data_models.keys())

        # Make sorted list of dates in all ensembles
        dates: Set[datetime.datetime] = set()
        for _, ens_data_model in data_models.items():
            dates = dates.union(ens_data_model.dates)
        sorted_dates: List[datetime.datetime] = sorted(list(dates))

        # Make list of all wells
        wells: List[str] = []
        for ens_data_model in data_models.values():
            wells.extend([well for well in ens_data_model.wells if well not in wells])

        self.add_settings_groups(
            {
                self.Ids.CHART_TYPE: WellOverviewChartType(),
                self.Ids.SELECTIONS: WellOverviewSelections(ensembles, sorted_dates),
                self.Ids.LAYOUT_OPTIONS: WellOverviewLayoutOptions(),
                self.Ids.FILTERS: WellOverviewFilters(wells),
            }
        )

        main_column = self.add_column()
        main_column.add_view_element(WellOverviewViewElement(), self.Ids.VIEW_ELEMENT)

    def set_callbacks(self) -> None:
        @callback(
            Output(
                {
                    "id": self.settings_group_unique_id(
                        self.Ids.LAYOUT_OPTIONS,
                        WellOverviewLayoutOptions.Ids.CHARTTYPE_SETTINGS,
                    ),
                    "charttype": ALL,
                },
                "style",
            ),
            Input(
                self.settings_group_unique_id(
                    self.Ids.CHART_TYPE, WellOverviewChartType.Ids.CHARTTYPE
                ),
                "value",
            ),
            State(
                {
                    "id": self.settings_group_unique_id(
                        self.Ids.LAYOUT_OPTIONS,
                        WellOverviewLayoutOptions.Ids.CHARTTYPE_SETTINGS,
                    ),
                    "charttype": ALL,
                },
                "id",
            ),
        )
        @callback_typecheck
        def _display_charttype_settings(
            chart_selected: ChartType, charttype_settings_ids: list
        ) -> list:
            """Display only the settings relevant for the currently selected chart type."""
            return [
                {"display": "block"}
                if settings_id["charttype"] == chart_selected
                else {"display": "none"}
                for settings_id in charttype_settings_ids
            ]

        @callback(
            Output(
                self.view_element(self.Ids.VIEW_ELEMENT)
                .component_unique_id(WellOverviewViewElement.Ids.GRAPH)
                .to_string(),
                "figure",
            ),
            Input(
                self.settings_group_unique_id(
                    self.Ids.SELECTIONS, WellOverviewSelections.Ids.ENSEMBLES
                ),
                "value",
            ),
            Input(
                {
                    "id": self.settings_group_unique_id(
                        self.Ids.LAYOUT_OPTIONS,
                        WellOverviewLayoutOptions.Ids.CHARTTYPE_CHECKLIST,
                    ),
                    "charttype": ALL,
                },
                "value",
            ),
            Input(
                self.settings_group_unique_id(
                    self.Ids.SELECTIONS, WellOverviewSelections.Ids.RESPONSE
                ),
                "value",
            ),
            Input(
                self.settings_group_unique_id(
                    self.Ids.SELECTIONS,
                    WellOverviewSelections.Ids.PROD_FROM_DATE,
                ),
                "value",
            ),
            Input(
                self.settings_group_unique_id(
                    self.Ids.SELECTIONS,
                    WellOverviewSelections.Ids.PROD_UNTIL_DATE,
                ),
                "value",
            ),
            Input(
                self.settings_group_unique_id(
                    self.Ids.CHART_TYPE, WellOverviewChartType.Ids.CHARTTYPE
                ),
                "value",
            ),
            Input(
                self.settings_group_unique_id(
                    self.Ids.FILTERS, WellOverviewFilters.Ids.SELECTED_WELLS
                ),
                "value",
            ),
            State(
                {
                    "id": self.settings_group_unique_id(
                        self.Ids.LAYOUT_OPTIONS,
                        WellOverviewLayoutOptions.Ids.CHARTTYPE_CHECKLIST,
                    ),
                    "charttype": ALL,
                },
                "id",
            ),
            State(
                self.view_element(self.Ids.VIEW_ELEMENT)
                .component_unique_id(WellOverviewViewElement.Ids.GRAPH)
                .to_string(),
                "figure",
            ),
        )
        @callback_typecheck
        def _update_graph(
            ensembles: List[str],
            checklist_values: List[List[str]],
            sumvec: str,
            prod_from_date: Union[str, None],
            prod_until_date: Union[str, None],
            charttype_selected: ChartType,
            wells_selected: List[str],
            checklist_ids: List[Dict[str, str]],
            current_fig_dict: Optional[Dict[str, Any]],
        ) -> Component:
            # pylint: disable=too-many-arguments
            """Updates the well overview graph with selected input (f.ex chart type)"""
            ctx = callback_context.triggered[0]["prop_id"].split(".")[0]

            settings = {
                checklist_id["charttype"]: checklist_values[i]
                for i, checklist_id in enumerate(checklist_ids)
            }

            # If the event is a plot settings event, then we only update the formatting
            # and not the figure data
            if (
                current_fig_dict is not None
                and self.settings_group(self.Ids.LAYOUT_OPTIONS)
                .component_unique_id(WellOverviewLayoutOptions.Ids.CHARTTYPE_CHECKLIST)
                .to_string()
                in ctx
            ):
                fig_dict = format_well_overview_figure(
                    figure=go.Figure(current_fig_dict),
                    charttype=charttype_selected,
                    settings=settings[charttype_selected],
                    sumvec=sumvec,
                    prod_from_date=prod_from_date,
                    prod_until_date=prod_until_date,
                )
            else:
                figure = WellOverviewFigure(
                    ensembles=ensembles,
                    data_models=self._data_models,
                    sumvec=sumvec,
                    prod_from_date=datetime.datetime.strptime(
                        prod_from_date, "%Y-%m-%d"
                    )
                    if prod_from_date is not None
                    else None,
                    prod_until_date=datetime.datetime.strptime(
                        prod_until_date, "%Y-%m-%d"
                    )
                    if prod_until_date is not None
                    else None,
                    charttype=charttype_selected,
                    wells=wells_selected,
                    theme=self._theme,
                )

                fig_dict = format_well_overview_figure(
                    figure=figure.figure,
                    charttype=charttype_selected,
                    settings=settings[charttype_selected],
                    sumvec=sumvec,
                    prod_from_date=prod_from_date,
                    prod_until_date=prod_until_date,
                )

            return fig_dict



# pylint: skip-file
import io
import json
import logging
from dataclasses import asdict
from typing import Dict, Optional, Union
from urllib.parse import quote_plus, unquote_plus
from uuid import uuid4

import flask
import flask_caching
from dash import Dash

from webviz_subsurface._utils.perf_timer import PerfTimer

from ._surface_to_image import surface_to_png_bytes
from .ensemble_surface_provider import (
    EnsembleSurfaceProvider,
    ObservedSurfaceAddress,
    SimulatedSurfaceAddress,
    StatisticalSurfaceAddress,
)

LOGGER = logging.getLogger(__name__)
ROOT_URL_PATH = "/SurfaceServerLazy"


class SurfaceServerLazy:
    def __init__(self, app: Dash) -> None:
        self._dash_app: Dash = app
        self._id_to_provider_dict: Dict[str, EnsembleSurfaceProvider] = {}

        self._image_cache = None
        # self._image_cache = flask_caching.Cache(
        #     config={
        #         "CACHE_TYPE": "RedisCache",
        #         "CACHE_KEY_PREFIX": f"SurfaceServer_{uuid4()}",
        #         "CACHE_REDIS_HOST": "localhost",
        #         "CACHE_REDIS_PORT": 6379,
        #         "CACHE_REDIS_URL": "redis://localhost:6379",
        #     }
        # )
        # self._image_cache = flask_caching.Cache(
        #     config={
        #         "CACHE_TYPE": "FileSystemCache",
        #         "CACHE_DIR": "/home/sigurdp/buf/flask_filesys_cache",
        #     }
        # )
        # self._image_cache.init_app(app.server)

    @staticmethod
    def instance(app: Dash) -> "SurfaceServerLazy":
        global SURFACE_SERVER_INSTANCE
        if not SURFACE_SERVER_INSTANCE:
            LOGGER.debug("Initializing SurfaceServerLazy instance")
            SURFACE_SERVER_INSTANCE = SurfaceServerLazy(app)

        return SURFACE_SERVER_INSTANCE

    def add_provider(self, provider: EnsembleSurfaceProvider) -> None:
        # Setup the url rule (our route) when the first provider is added
        if not self._id_to_provider_dict:
            self._setup_url_rule()

        provider_id = provider.provider_id()
        LOGGER.debug(f"Adding provider with id={provider_id}")

        existing_provider = self._id_to_provider_dict.get(provider_id)
        if existing_provider:
            # Issue a warning if there already is a provider registered with the same
            # id AND if the actual provider instance is different.
            # This should not be a problem, but will happen until the provider factory
            # gets caching.
            if existing_provider is not provider:
                LOGGER.warning(
                    f"Provider with id={provider_id} ignored, the id is already present"
                )
                return

        self._id_to_provider_dict[provider_id] = provider

        # routes = []
        # for rule in self._dash_app.server.url_map.iter_rules():
        #     routes.append("%s" % rule)

        # for route in routes:
        #     print(route)

    def encode_partial_url(
        self,
        provider_id: str,
        address: Union[
            StatisticalSurfaceAddress, SimulatedSurfaceAddress, ObservedSurfaceAddress
        ],
    ) -> str:
        if not provider_id in self._id_to_provider_dict:
            raise ValueError("Could not find provider")

        if isinstance(address, StatisticalSurfaceAddress):
            addr_type_str = "sta"
        elif isinstance(address, SimulatedSurfaceAddress):
            addr_type_str = "sim"
        elif isinstance(address, ObservedSurfaceAddress):
            addr_type_str = "obs"

        surf_address_str = quote_plus(json.dumps(asdict(address)))

        url_path: str = (
            f"{ROOT_URL_PATH}/{provider_id}/{addr_type_str}/{surf_address_str}"
        )
        return url_path

    def _setup_url_rule(self) -> None:
        @self._dash_app.server.route(
            ROOT_URL_PATH + "/<provider_id>/<addr_type_str>/<surf_address_str>"
        )
        def _handle_request(
            provider_id: str, addr_type_str: str, surf_address_str: str
        ) -> flask.Response:
            LOGGER.debug(
                f"Handling request: "
                f"provider_id={provider_id} "
                f"addr_type_str={addr_type_str} "
                f"surf_address_str={surf_address_str}"
            )

            timer = PerfTimer()

            try:
                provider = self._id_to_provider_dict[provider_id]
                surf_address_dict = json.loads(unquote_plus(surf_address_str))
                address: Union[
                    StatisticalSurfaceAddress,
                    SimulatedSurfaceAddress,
                    ObservedSurfaceAddress,
                ]
                if addr_type_str == "sta":
                    address = StatisticalSurfaceAddress(**surf_address_dict)
                if addr_type_str == "sim":
                    address = SimulatedSurfaceAddress(**surf_address_dict)
                if addr_type_str == "obs":
                    address = ObservedSurfaceAddress(**surf_address_dict)
            except:
                LOGGER.error("Error decoding surface address")
                flask.abort(404)

            if self._image_cache:
                img_cache_key = (
                    f"provider_id={provider_id} "
                    f"addr_type={addr_type_str} address={surf_address_str}"
                )
                LOGGER.debug(
                    f"Looking for image in cache (key={img_cache_key}, "
                    f"cache_type={self._image_cache.config['CACHE_TYPE']})"
                )
                cached_img_bytes = self._image_cache.get(img_cache_key)
                if cached_img_bytes:
                    response = flask.send_file(
                        io.BytesIO(cached_img_bytes), mimetype="image/png"
                    )
                    LOGGER.debug(
                        f"Request handled from image cache in: {timer.elapsed_s():.2f}s"
                    )
                    return response

            LOGGER.debug("Getting surface from provider...")
            timer.lap_s()
            surface = provider.get_surface(address)
            if not surface:
                LOGGER.error(f"Error getting surface for address: {address}")
                flask.abort(404)
            et_get_s = timer.lap_s()
            LOGGER.debug(
                f"Got surface (dimensions={surface.dimensions}, #cells={surface.ncol*surface.nrow})"
            )

            LOGGER.debug("Converting to PNG image...")
            png_bytes: bytes = surface_to_png_bytes(surface)
            LOGGER.debug(
                f"Got PNG image, size={(len(png_bytes) / (1024 * 1024)):.2f}MB"
            )
            et_to_image_s = timer.lap_s()

            LOGGER.debug("Sending image")
            response = flask.send_file(io.BytesIO(png_bytes), mimetype="image/png")
            et_send_s = timer.lap_s()

            if self._image_cache and img_cache_key:
                self._image_cache.add(img_cache_key, png_bytes)

            LOGGER.debug(
                f"Request handled in: {timer.elapsed_s():.2f}s ("
                f"get={et_get_s:.2f}s, to_image={et_to_image_s:.2f}s, send={et_send_s:.2f}s)"
            )

            return response


SURFACE_SERVER_INSTANCE: Optional[SurfaceServerLazy] = None



import warnings

# pylint: disable=no-name-in-module
from webviz_config.plugins import ReservoirSimulationTimeSeries


def test_reservoir_simulation_timeseries(
    dash_duo, app, shared_settings, testdata_folder
) -> None:
    plugin = ReservoirSimulationTimeSeries(
        app,
        shared_settings["HM_SETTINGS"],
        ensembles=shared_settings["HM_ENSEMBLES"],
        obsfile=testdata_folder
        / "01_drogon_ahm"
        / "share"
        / "observations"
        / "tables"
        / "ert_observations.yml",
    )
    app.layout = plugin.layout
    dash_duo.start_server(app)

    logs = []
    for log in dash_duo.get_logs():
        if "dash_renderer" in log.get("message"):
            warnings.warn(log.get("message"))
        else:
            logs.append(log)
    assert not logs



from .clientside_stores import clientside_stores
from .intersection_and_map import intersection_and_map_layout
from .intersection_data import intersection_data_layout
from .map_data import map_data_layout
from .realization_filter import realization_layout
from .uncertainty_table import uncertainty_table_layout



import json
from typing import Any, Callable, Dict, List

import webviz_subsurface_components as wsc
from dash import Dash, Input, Output

from ._business_logic import WellCompletionsDataModel
from ._layout import LayoutElements


def plugin_callbacks(
    app: Dash, get_uuid: Callable, data_models: Dict[str, WellCompletionsDataModel]
) -> None:
    @app.callback(
        Output(get_uuid(LayoutElements.WELL_COMPLETIONS_COMPONENT), "children"),
        Output(get_uuid(LayoutElements.WELL_COMPLETIONS_COMPONENT), "style"),
        Input(get_uuid(LayoutElements.ENSEMBLE_DROPDOWN), "value"),
    )
    def _render_well_completions(ensemble_name: str) -> list:
        data = json.load(data_models[ensemble_name].create_ensemble_dataset())

        no_leaves = count_leaves(data["stratigraphy"])
        return [
            wsc.WellCompletions(id="well_completions", data=data),
            {
                "padding": "10px",
                "height": no_leaves * 50 + 180,
                "min-height": 500,
                "width": "98%",
            },
        ]


def count_leaves(stratigraphy: List[Dict[str, Any]]) -> int:
    """Counts the number of leaves in the stratigraphy tree"""
    return sum(
        count_leaves(zonedict["subzones"]) if "subzones" in zonedict else 1
        for zonedict in stratigraphy
    )



# pylint: disable=no-name-in-module
from webviz_config.plugins import VfpAnalysis
from webviz_config.testing import WebvizComposite


def test_vfp_analysis(_webviz_duo: WebvizComposite, shared_settings: dict) -> None:
    plugin = VfpAnalysis(
        shared_settings["HM_SETTINGS"], vfp_file_pattern="tests/data/vfp.arrow"
    )

    _webviz_duo.start_server(plugin)

    assert _webviz_duo.get_logs() == []



import datetime
from typing import List, Optional, Sequence

import pandas as pd
from webviz_subsurface_components import ExpressionInfo

from webviz_subsurface._providers import EnsembleSummaryProvider, Frequency
from webviz_subsurface._utils.vector_calculator import (
    create_calculated_vector_df,
    get_selected_expressions,
)

from .. import dataframe_utils
from ..from_timeseries_cumulatives import (
    calculate_from_resampled_cumulative_vectors_df,
    get_cumulative_vector_name,
    is_per_interval_or_per_day_vector,
)
from .derived_vectors_accessor import DerivedVectorsAccessor


class DerivedEnsembleVectorsAccessorImpl(DerivedVectorsAccessor):
    """
    Class to create derived vector data and access these for a regular ensemble.

    The ensemble is represented with an ensemble summary provider.

    A list of vector names are provided, and data is fetched or created based on which
    type of vectors are present in the list.

    Vector names can be regular vectors existing among vector names in the provider, Per
    Interval/Per Day vector or a calculated vector from vector calculator.

    Based on the vector type, the class provides an interface for retrieving dataframes
    for the set of such vectors for the provider.
    """

    def __init__(
        self,
        name: str,
        provider: EnsembleSummaryProvider,
        vectors: List[str],
        expressions: Optional[List[ExpressionInfo]] = None,
        resampling_frequency: Optional[Frequency] = None,
        relative_date: Optional[datetime.datetime] = None,
    ) -> None:
        # Initialize base class
        super().__init__(provider.realizations())

        self._name = name
        self._provider = provider
        self._provider_vectors = [
            vector for vector in vectors if vector in self._provider.vector_names()
        ]
        self._per_interval_and_per_day_vectors = [
            vector
            for vector in vectors
            if is_per_interval_or_per_day_vector(vector)
            and get_cumulative_vector_name(vector) in provider.vector_names()
        ]
        self._vector_calculator_expressions = (
            get_selected_expressions(expressions, vectors)
            if expressions is not None
            else []
        )
        self._resampling_frequency = (
            resampling_frequency if self._provider.supports_resampling() else None
        )
        self._relative_date = relative_date

    def has_provider_vectors(self) -> bool:
        return len(self._provider_vectors) > 0

    def has_per_interval_and_per_day_vectors(self) -> bool:
        return len(self._per_interval_and_per_day_vectors) > 0

    def has_vector_calculator_expressions(self) -> bool:
        return len(self._vector_calculator_expressions) > 0

    def get_provider_vectors_df(
        self, realizations: Optional[Sequence[int]] = None
    ) -> pd.DataFrame:
        """Get dataframe for the selected provider vectors"""
        if not self.has_provider_vectors():
            raise ValueError(
                f'Vector data handler for provider "{self._name}" has no provider vectors'
            )

        if self._relative_date:
            return dataframe_utils.create_relative_to_date_df(
                self._provider.get_vectors_df(
                    self._provider_vectors, self._resampling_frequency, realizations
                ),
                self._relative_date,
            )
        return self._provider.get_vectors_df(
            self._provider_vectors, self._resampling_frequency, realizations
        )

    def create_per_interval_and_per_day_vectors_df(
        self,
        realizations: Optional[Sequence[int]] = None,
    ) -> pd.DataFrame:
        """Get dataframe with interval delta and per day delta vector data for provided vectors.

        The returned dataframe contains columns with name of vector and corresponding interval delta
        or per day delta data.

        Interval delta and per day delta date is calculated with same sampling frequency as provider
        is set with. I.e. resampling frequency is given for providers supporting resampling,
        otherwise sampling frequency is fixed.

        `Input:`
        * realizations: Sequence[int] - Sequency of realization numbers to include in calculation

        `Output:`
        * dataframe with interval  vector names in columns and their cumulative data in rows.
        `Columns` in dataframe: ["DATE", "REAL", vector1, ..., vectorN]

        ---------------------
        `NOTE:`
        * Handle calculation of cumulative when raw data is added
        * See TODO in calculate_from_resampled_cumulative_vectors_df()
        """
        if not self.has_per_interval_and_per_day_vectors():
            raise ValueError(
                f'Vector data handler for provider "{self._name}" has no per interval '
                "or per day vector names"
            )

        cumulative_vector_names = [
            get_cumulative_vector_name(elm)
            for elm in self._per_interval_and_per_day_vectors
            if is_per_interval_or_per_day_vector(elm)
        ]
        cumulative_vector_names = list(sorted(set(cumulative_vector_names)))

        vectors_df = self._provider.get_vectors_df(
            cumulative_vector_names, self._resampling_frequency, realizations
        )

        per_interval_and_per_day_vectors_df = pd.DataFrame()
        for vector_name in self._per_interval_and_per_day_vectors:
            cumulative_vector_name = get_cumulative_vector_name(vector_name)
            per_interval_or_per_day_vector_df = (
                calculate_from_resampled_cumulative_vectors_df(
                    vectors_df[["DATE", "REAL", cumulative_vector_name]],
                    as_per_day=vector_name.startswith("PER_DAY_"),
                )
            )
            if per_interval_and_per_day_vectors_df.empty:
                per_interval_and_per_day_vectors_df = per_interval_or_per_day_vector_df
            else:
                per_interval_and_per_day_vectors_df = pd.merge(
                    per_interval_and_per_day_vectors_df,
                    per_interval_or_per_day_vector_df,
                    how="inner",
                )

        if self._relative_date:
            return dataframe_utils.create_relative_to_date_df(
                per_interval_and_per_day_vectors_df,
                self._relative_date,
            )
        return per_interval_and_per_day_vectors_df

    def create_calculated_vectors_df(
        self, realizations: Optional[Sequence[int]] = None
    ) -> pd.DataFrame:
        """Get dataframe with calculated vector data for provided vectors.

        The returned dataframe contains columns with name of vector and corresponding calculated
        data.

        Calculated vectors are created with same sampling frequency as provider is set with. I.e.
        resampling frequency is given for providers supporting resampling, otherwise sampling
        frequency is fixed.

        `Input:`
        * realizations: Sequence[int] - Sequency of realization numbers to include in calculation

        `Output:`
        * dataframe with vector names in columns and their calculated data in rows
        `Columns` in dataframe: ["DATE", "REAL", vector1, ..., vectorN]
        """
        if not self.has_vector_calculator_expressions():
            raise ValueError(
                f'Assembled vector data accessor for provider "{self._name}"'
                "has no vector calculator expressions"
            )
        calculated_vectors_df = pd.DataFrame()
        for expression in self._vector_calculator_expressions:
            calculated_vector_df = create_calculated_vector_df(
                expression, self._provider, realizations, self._resampling_frequency
            )
            if calculated_vectors_df.empty:
                calculated_vectors_df = calculated_vector_df
            else:
                calculated_vectors_df = pd.merge(
                    calculated_vectors_df,
                    calculated_vector_df,
                    how="inner",
                )

        if self._relative_date:
            return dataframe_utils.create_relative_to_date_df(
                calculated_vectors_df,
                self._relative_date,
            )
        return calculated_vectors_df



from enum import Enum


class LayerTypes(str, Enum):
    HILLSHADING = "Hillshading2DLayer"
    MAP3D = "MapLayer"
    COLORMAP = "ColormapLayer"
    WELL = "WellsLayer"
    WELLTOPSLAYER = "GeoJsonLayer"
    DRAWING = "DrawingLayer"
    FAULTPOLYGONS = "FaultPolygonsLayer"
    GEOJSON = "GeoJsonLayer"


class LayerNames(str, Enum):
    HILLSHADING = "Surface (hillshading)"
    MAP3D = "3D Map"
    COLORMAP = "Surface (color)"
    WELL = "Wells"
    WELLTOPSLAYER = "Well tops"
    DRAWING = "Drawings"
    FAULTPOLYGONS = "Fault polygons"
    GEOJSON = "GeoJsonLayer"


class SurfaceMode(str, Enum):
    MEAN = "Mean"
    REALIZATION = "Single realization"
    OBSERVED = "Observed"
    STDDEV = "StdDev"
    MINIMUM = "Minimum"
    MAXIMUM = "Maximum"
    P10 = "P10"
    P90 = "P90"



import json
from typing import Dict

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc

_MAIN_WEBVIZ_METADATA_KEY = b"webviz"
_PER_VECTOR_MIN_MAX_KEY = "per_vector_min_max"


def find_min_max_for_numeric_table_columns(
    table: pa.Table,
) -> Dict[str, dict]:
    """Determine per-vector min/max values and return as dictionary indexed by vector name"""
    ret_dict = {}
    for field in table.schema:
        if pa.types.is_floating(field.type):
            # pylint: disable=no-member
            minmax = pc.min_max(table[field.name])
            ret_dict[field.name] = {
                "min": minmax.get("min").as_py(),
                "max": minmax.get("max").as_py(),
            }

    return ret_dict


def add_per_vector_min_max_to_table_schema_metadata(
    table: pa.Table, per_vector_min_max: Dict[str, dict]
) -> pa.Table:
    """Store dict with per-vector min/max values schema's metadata"""

    webviz_meta = {_PER_VECTOR_MIN_MAX_KEY: per_vector_min_max}
    new_combined_meta = {}
    if table.schema.metadata is not None:
        new_combined_meta.update(table.schema.metadata)
    new_combined_meta.update({_MAIN_WEBVIZ_METADATA_KEY: json.dumps(webviz_meta)})
    table = table.replace_schema_metadata(new_combined_meta)
    return table


def get_per_vector_min_max_from_schema_metadata(schema: pa.Schema) -> Dict[str, dict]:
    """Extract dict containing per-vector min/max values from the schema-level metadata"""

    webviz_meta = json.loads(schema.metadata[_MAIN_WEBVIZ_METADATA_KEY])
    return webviz_meta[_PER_VECTOR_MIN_MAX_KEY]


def find_intersected_dates_between_realizations(table: pa.Table) -> np.ndarray:
    """Find the intersection of dates present in all the realizations
    The input table must contain both REAL and DATE columns, but this function makes
    no assumptions about sorting of either column"""

    unique_reals = table.column("REAL").unique().to_numpy()

    date_intersection = None
    for real in unique_reals:
        # pylint: disable=no-member
        real_mask = pc.is_in(table["REAL"], value_set=pa.array([real]))
        dates_in_real = table.filter(real_mask).column("DATE").unique().to_numpy()
        if date_intersection is None:
            date_intersection = dates_in_real
        else:
            date_intersection = np.intersect1d(
                date_intersection, dates_in_real, assume_unique=True
            )

    if date_intersection is not None:
        return date_intersection

    return np.empty(0, dtype=np.datetime64)



########################################
#
#  Copyright (C) 2021-     Equinor ASA
#
#  webviz-subsurface is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  webviz-subsurface is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http:#www.gnu.org/licenses/gpl.html>
#  for more details.
#
########################################
########################################
#
#  The code in this file is based on and/or inspired by opm-common,
#  which is distributed under the GNU General Public License v3.0,
#  and available at https://github.com/OPM/opm-common.
#
#  Especially in the following points this file, or files imported in this file
#  within the same context, make use of or are based on methods developed
#  for the opm-common project:
#  - Reading of Eclipse INIT files
#  - Units, unit systems and conversion of units
#  - Class structures for storing and accessing PVT data,
#    both base classes and derived classes for specific fluids
#
#  You should have received a copy of the GNU General Public License
#  along with webviz-subsurface. If not, see (https://www.gnu.org/licenses/gpl-3.0.html).
#
#  (12-01-2021)
#
#  Copyright (C) 2009-2012 SINTEF ICT, Applied Mathematics.
#
########################################

from enum import IntEnum
from typing import Callable, List

from .units import Prefix, Unit, UnitBase


class EclUnitEnum(IntEnum):
    """An enum for the different unit systems"""

    ECL_SI_UNITS = 0
    ECL_METRIC_UNITS = 1
    ECL_FIELD_UNITS = 2
    ECL_LAB_UNITS = 3
    ECL_PVT_M_UNITS = 4


def unit_system_name(unit_system: EclUnitEnum) -> str:
    """Returns a string representation of the given unit system enum.

    Args:
        unit_system: Representation of the unit system as an enum.

    Returns:
        A string representation of the given unit system. In case of
        an undefined unit system, returns "UNKNOWN".

    """
    if unit_system == EclUnitEnum.ECL_SI_UNITS:
        return "SI"
    if unit_system == EclUnitEnum.ECL_METRIC_UNITS:
        return "METRIC"
    if unit_system == EclUnitEnum.ECL_FIELD_UNITS:
        return "FIELD"
    if unit_system == EclUnitEnum.ECL_LAB_UNITS:
        return "LAB"
    if unit_system == EclUnitEnum.ECL_PVT_M_UNITS:
        return "PVT-M"
    return "UNKNOWN"


class UnitSystems:
    # pylint: disable=too-few-public-methods
    """Namespace for unit systems"""

    class SI:
        # pylint: disable=too-few-public-methods
        """Namespace for SI unit system"""

        Pressure = Unit.pascal
        Temperature = Unit.deg_kelvin
        TemperatureOffset = 0.0
        AbsoluteTemperature = Unit.deg_kelvin
        Length = Unit.meter
        Time = Unit.second
        Mass = Unit.kilogram
        Permeability = Unit.square(Unit.meter)
        Transmissibility = Unit.cubic(Unit.meter)
        LiquidSurfaceVolume = Unit.Base(Unit.cubic(Unit.meter), "Sm^3")
        GasSurfaceVolume = Unit.Base(Unit.cubic(Unit.meter), "Sm^3")
        ReservoirVolume = Unit.Base(Unit.cubic(Unit.meter), "Rm^3")
        GasDissolutionFactor = GasSurfaceVolume / LiquidSurfaceVolume
        OilDissolutionFactor = LiquidSurfaceVolume / GasSurfaceVolume
        Density = Unit.kilogram / Unit.cubic(Unit.meter)
        PolymerDensity = Unit.kilogram / Unit.cubic(Unit.meter)
        Salinity = Unit.kilogram / Unit.cubic(Unit.meter)
        Viscosity = Unit.pascal * Unit.second
        Timestep = Unit.second
        SurfaceTension = Unit.newton / Unit.meter
        Energy = Unit.joule

    class Metric:
        # pylint: disable=too-few-public-methods
        """Namespace for metric unit system"""

        Pressure = Unit.bar
        Temperature = Unit.deg_celsius
        TemperatureOffset = Unit.deg_celsius_offset
        AbsoluteTemperature = Unit.deg_celsius
        Length = Unit.meter
        Time = Unit.day
        Mass = Unit.kilogram
        Permeability = Prefix.milli * Unit.darcy
        Transmissibility = (
            Prefix.centi * Unit.poise * Unit.cubic(Unit.meter) / (Unit.day * Unit.bar)
        )
        LiquidSurfaceVolume = Unit.Base(Unit.cubic(Unit.meter), "Sm^3")
        GasSurfaceVolume = Unit.Base(Unit.cubic(Unit.meter), "Sm^3")
        ReservoirVolume = Unit.Base(Unit.cubic(Unit.meter), "Rm^3")
        GasDissolutionFactor = GasSurfaceVolume / LiquidSurfaceVolume
        OilDissolutionFactor = LiquidSurfaceVolume / GasSurfaceVolume
        Density = Unit.kilogram / Unit.cubic(Unit.meter)
        PolymerDensity = Unit.kilogram / Unit.cubic(Unit.meter)
        Salinity = Unit.kilogram / Unit.cubic(Unit.meter)
        Viscosity = Prefix.centi * Unit.poise
        Timestep = Unit.day
        SurfaceTension = Unit.dyne / (Prefix.centi * Unit.meter)
        Energy = Prefix.kilo * Unit.joule

    class Field:
        # pylint: disable=too-few-public-methods
        """Namespace for field unit system"""

        Pressure = Unit.psi
        Temperature = Unit.deg_fahrenheit
        TemperatureOffset = Unit.deg_fahrenheit_offset
        AbsoluteTemperature = Unit.deg_fahrenheit
        Length = Unit.feet
        Time = Unit.day
        Mass = Unit.pound
        Permeability = Prefix.milli * Unit.darcy
        Transmissibility = Prefix.centi * Unit.poise * Unit.stb / (Unit.day * Unit.psi)
        LiquidSurfaceVolume = Unit.Base(Unit.barrel, "rstb")
        GasSurfaceVolume = Unit.Base(Unit.cubic(Unit.feet), "scf")
        ReservoirVolume = Unit.Base(Unit.barrel, "RB")
        GasDissolutionFactor = GasSurfaceVolume / LiquidSurfaceVolume
        OilDissolutionFactor = LiquidSurfaceVolume / GasSurfaceVolume
        Density = Unit.pound / Unit.cubic(Unit.feet)
        PolymerDensity = Unit.pound / Unit.stb
        Salinity = Unit.pound / Unit.stb
        Viscosity = Prefix.centi * Unit.poise
        Timestep = Unit.day
        SurfaceTension = Unit.dyne / (Prefix.centi * Unit.meter)
        Energy = Unit.btu

    class Lab:
        # pylint: disable=too-few-public-methods
        """Namespace for lab unit system"""

        Pressure = Unit.atm
        Temperature = Unit.deg_celsius
        TemperatureOffset = Unit.deg_celsius_offset
        AbsoluteTemperature = Unit.deg_celsius
        Length = Prefix.centi * Unit.meter
        Time = Unit.hour
        Mass = Unit.gram
        Permeability = Prefix.milli * Unit.darcy
        Transmissibility = (
            Prefix.centi
            * Unit.poise
            * Unit.cubic(Prefix.centi * Unit.meter)
            / (Unit.hour * Unit.atm)
        )
        LiquidSurfaceVolume = Unit.Base(Unit.cubic(Prefix.centi * Unit.meter), "Scm^3")
        GasSurfaceVolume = Unit.Base(Unit.cubic(Prefix.centi * Unit.meter), "Scm^3")
        ReservoirVolume = Unit.Base(Unit.cubic(Prefix.centi * Unit.meter), "Rcm^3")
        GasDissolutionFactor = GasSurfaceVolume / LiquidSurfaceVolume
        OilDissolutionFactor = LiquidSurfaceVolume / GasSurfaceVolume
        Density = Unit.gram / Unit.cubic(Prefix.centi * Unit.meter)
        PolymerDensity = Unit.gram / Unit.cubic(Prefix.centi * Unit.meter)
        Salinity = Unit.gram / Unit.cubic(Prefix.centi * Unit.meter)
        Viscosity = Prefix.centi * Unit.poise
        Timestep = Unit.hour
        SurfaceTension = Unit.dyne / (Prefix.centi * Unit.meter)
        Energy = Unit.joule

    class PVTM:
        # pylint: disable=too-few-public-methods
        """Namespace for PVTM unit system"""

        Pressure = Unit.atm
        Temperature = Unit.deg_celsius
        TemperatureOffset = Unit.deg_celsius_offset
        AbsoluteTemperature = Unit.deg_celsius
        Length = Unit.meter
        Time = Unit.day
        Mass = Unit.kilogram
        Permeability = Prefix.milli * Unit.darcy
        Transmissibility = (
            Prefix.centi * Unit.poise * Unit.cubic(Unit.meter) / (Unit.day * Unit.atm)
        )
        LiquidSurfaceVolume = Unit.Base(Unit.cubic(Unit.meter), "Sm^3")
        GasSurfaceVolume = Unit.Base(Unit.cubic(Unit.meter), "Sm^3")
        ReservoirVolume = Unit.Base(Unit.cubic(Unit.meter), "Rm^3")
        GasDissolutionFactor = GasSurfaceVolume / LiquidSurfaceVolume
        OilDissolutionFactor = LiquidSurfaceVolume / GasSurfaceVolume
        Density = Unit.kilogram / Unit.cubic(Unit.meter)
        PolymerDensity = Unit.kilogram / Unit.cubic(Unit.meter)
        Salinity = Unit.kilogram / Unit.cubic(Unit.meter)
        Viscosity = Prefix.centi * Unit.poise
        Timestep = Unit.day
        SurfaceTension = Unit.dyne / (Prefix.centi * Unit.meter)
        Energy = Prefix.kilo * Unit.joule


class EclUnits:
    """Namespace for units used in Eclipse"""

    class UnitSystem:
        """Base interface class for Eclipse specific unit systems.

        Raises NotImplementedErrors when used directly.
        """

        @staticmethod
        def density() -> UnitBase:
            """
            Returns:
                The unit system's unit of density (e.g. 1 kg/m^3)
            """
            raise NotImplementedError("The base class cannot be called directly.")

        @staticmethod
        def depth() -> UnitBase:
            """
            Returns:
                The unit system's unit of depth (e.g. 1 m)
            """
            raise NotImplementedError("The base class cannot be called directly.")

        @staticmethod
        def pressure() -> UnitBase:
            """
            Returns:
                The unit system's unit of pressure (e.g. 1 Pa)
            """
            raise NotImplementedError("The base class cannot be called directly.")

        @staticmethod
        def reservoir_rate() -> UnitBase:
            """
            Returns:
                The unit system's unit of the reservoir rate (e.g. m^3/s)
            """
            raise NotImplementedError("The base class cannot be called directly.")

        @staticmethod
        def reservoir_volume() -> UnitBase:
            """
            Returns:
                The unit system's unit of reservoir volume (e.g. rm^3)
            """
            raise NotImplementedError("The base class cannot be called directly.")

        @staticmethod
        def surface_volume_gas() -> UnitBase:
            """
            Returns:
                The unit system's unit of the surface volume of gas (e.g. sm^3)
            """
            raise NotImplementedError("The base class cannot be called directly.")

        @staticmethod
        def surface_volume_liquid() -> UnitBase:
            """
            Returns:
                The unit system's unit of the surface volume of liquids (e.g. sm^3)
            """
            raise NotImplementedError("The base class cannot be called directly.")

        @staticmethod
        def time() -> UnitBase:
            """
            Returns:
                The unit system's unit of time (e.g. s)
            """
            raise NotImplementedError("The base class cannot be called directly.")

        @staticmethod
        def transmissibility() -> UnitBase:
            """
            Returns:
                The unit system's unit of transmissibility (e.g. m^3)
            """
            raise NotImplementedError("The base class cannot be called directly.")

        @staticmethod
        def viscosity() -> UnitBase:
            """
            Returns:
                The unit system's unit of viscosity (e.g. Pa*s)
            """
            raise NotImplementedError("The base class cannot be called directly.")

        def dissolved_gas_oil_ratio(self) -> UnitBase:
            """
            Returns:
                The unit system's unit of the dissolved gas to oil ratio (e.g. [-])
            """
            return self.surface_volume_gas() / self.surface_volume_liquid()

        def vaporised_oil_gas_ratio(self) -> UnitBase:
            """
            Returns:
                The unit system's unit of the vaporised oil to gas ratio (e.g. [-])
            """
            return self.surface_volume_liquid() / self.surface_volume_gas()

    class EclSIUnitSystem(UnitSystem):
        @staticmethod
        def density() -> UnitBase:
            return UnitSystems.SI.Density

        @staticmethod
        def depth() -> UnitBase:
            return UnitSystems.SI.Length

        @staticmethod
        def pressure() -> UnitBase:
            return UnitSystems.SI.Pressure

        @staticmethod
        def reservoir_rate() -> UnitBase:
            return UnitSystems.SI.ReservoirVolume / UnitSystems.SI.Time

        @staticmethod
        def reservoir_volume() -> UnitBase:
            return UnitSystems.SI.ReservoirVolume

        @staticmethod
        def surface_volume_gas() -> UnitBase:
            return UnitSystems.SI.GasSurfaceVolume

        @staticmethod
        def surface_volume_liquid() -> UnitBase:
            return UnitSystems.SI.LiquidSurfaceVolume

        @staticmethod
        def time() -> UnitBase:
            return UnitSystems.SI.Time

        @staticmethod
        def transmissibility() -> UnitBase:
            return UnitSystems.SI.Transmissibility

        @staticmethod
        def viscosity() -> UnitBase:
            return UnitSystems.SI.Viscosity

    class EclMetricUnitSystem(UnitSystem):
        @staticmethod
        def density() -> UnitBase:
            return UnitSystems.Metric.Density

        @staticmethod
        def depth() -> UnitBase:
            return UnitSystems.Metric.Length

        @staticmethod
        def pressure() -> UnitBase:
            return UnitSystems.Metric.Pressure

        @staticmethod
        def reservoir_rate() -> UnitBase:
            return UnitSystems.Metric.ReservoirVolume / UnitSystems.Metric.Time

        @staticmethod
        def reservoir_volume() -> UnitBase:
            return UnitSystems.Metric.ReservoirVolume

        @staticmethod
        def surface_volume_gas() -> UnitBase:
            return UnitSystems.Metric.GasSurfaceVolume

        @staticmethod
        def surface_volume_liquid() -> UnitBase:
            return UnitSystems.Metric.LiquidSurfaceVolume

        @staticmethod
        def time() -> UnitBase:
            return UnitSystems.Metric.Time

        @staticmethod
        def transmissibility() -> UnitBase:
            return UnitSystems.Metric.Transmissibility

        @staticmethod
        def viscosity() -> UnitBase:
            return UnitSystems.Metric.Viscosity

    class EclFieldUnitSystem(UnitSystem):
        @staticmethod
        def density() -> UnitBase:
            return UnitSystems.Field.Density

        @staticmethod
        def depth() -> UnitBase:
            return UnitSystems.Field.Length

        @staticmethod
        def pressure() -> UnitBase:
            return UnitSystems.Field.Pressure

        @staticmethod
        def reservoir_rate() -> UnitBase:
            return UnitSystems.Field.ReservoirVolume / UnitSystems.Field.Time

        @staticmethod
        def reservoir_volume() -> UnitBase:
            return UnitSystems.Field.ReservoirVolume

        @staticmethod
        def surface_volume_gas() -> UnitBase:
            return UnitSystems.Field.GasSurfaceVolume

        @staticmethod
        def surface_volume_liquid() -> UnitBase:
            return UnitSystems.Field.LiquidSurfaceVolume

        @staticmethod
        def time() -> UnitBase:
            return UnitSystems.Field.Time

        @staticmethod
        def transmissibility() -> UnitBase:
            return UnitSystems.Field.Transmissibility

        @staticmethod
        def viscosity() -> UnitBase:
            return UnitSystems.Field.Viscosity

    class EclLabUnitSystem(UnitSystem):
        @staticmethod
        def density() -> UnitBase:
            return UnitSystems.Lab.Density

        @staticmethod
        def depth() -> UnitBase:
            return UnitSystems.Lab.Length

        @staticmethod
        def pressure() -> UnitBase:
            return UnitSystems.Lab.Pressure

        @staticmethod
        def reservoir_rate() -> UnitBase:
            return UnitSystems.Lab.ReservoirVolume / UnitSystems.Lab.Time

        @staticmethod
        def reservoir_volume() -> UnitBase:
            return UnitSystems.Lab.ReservoirVolume

        @staticmethod
        def surface_volume_gas() -> UnitBase:
            return UnitSystems.Lab.GasSurfaceVolume

        @staticmethod
        def surface_volume_liquid() -> UnitBase:
            return UnitSystems.Lab.LiquidSurfaceVolume

        @staticmethod
        def time() -> UnitBase:
            return UnitSystems.Lab.Time

        @staticmethod
        def transmissibility() -> UnitBase:
            return UnitSystems.Lab.Transmissibility

        @staticmethod
        def viscosity() -> UnitBase:
            return UnitSystems.Lab.Viscosity

    class EclPvtMUnitSystem(UnitSystem):
        @staticmethod
        def density() -> UnitBase:
            return UnitSystems.PVTM.Density

        @staticmethod
        def depth() -> UnitBase:
            return UnitSystems.PVTM.Length

        @staticmethod
        def pressure() -> UnitBase:
            return UnitSystems.PVTM.Pressure

        @staticmethod
        def reservoir_rate() -> UnitBase:
            return UnitSystems.PVTM.ReservoirVolume / UnitSystems.PVTM.Time

        @staticmethod
        def reservoir_volume() -> UnitBase:
            return UnitSystems.PVTM.ReservoirVolume

        @staticmethod
        def surface_volume_gas() -> UnitBase:
            return UnitSystems.PVTM.GasSurfaceVolume

        @staticmethod
        def surface_volume_liquid() -> UnitBase:
            return UnitSystems.PVTM.LiquidSurfaceVolume

        @staticmethod
        def time() -> UnitBase:
            return UnitSystems.PVTM.Time

        @staticmethod
        def transmissibility() -> UnitBase:
            return UnitSystems.PVTM.Transmissibility

        @staticmethod
        def viscosity() -> UnitBase:
            return UnitSystems.PVTM.Viscosity

    @staticmethod
    def create_unit_system(unit_system: int) -> UnitSystem:
        if unit_system == EclUnitEnum.ECL_SI_UNITS:
            return EclUnits.EclSIUnitSystem()
        if unit_system == EclUnitEnum.ECL_METRIC_UNITS:
            return EclUnits.EclMetricUnitSystem()
        if unit_system == EclUnitEnum.ECL_FIELD_UNITS:
            return EclUnits.EclFieldUnitSystem()
        if unit_system == EclUnitEnum.ECL_LAB_UNITS:
            return EclUnits.EclLabUnitSystem()
        if unit_system == EclUnitEnum.ECL_PVT_M_UNITS:
            return EclUnits.EclPvtMUnitSystem()
        raise ValueError(f"Unsupported Unit Convention: {unit_system}")


class ConvertUnits:
    # pylint: disable=too-few-public-methods
    """A structure holding callables for unit conversions.

    Attributes:
        independent: Callable for unit conversion of independent variable.
        column: List of callables for unit conversion of each of the dependent variables.
    """

    def __init__(
        self,
        independent: Callable[
            [
                float,
            ],
            float,
        ],
        column: List[
            Callable[
                [
                    float,
                ],
                float,
            ]
        ],
    ) -> None:
        """Initiates an object holding conversion callables
        for each of the independent and dependent variables
        respectively.

        Args:
            independent: Callable for unit conversion of independent variable.
            column: List of callables for unit conversion of each of the dependent variables.
        """
        self.independent: Callable[
            [
                float,
            ],
            float,
        ] = independent
        self.column: List[
            Callable[
                [
                    float,
                ],
                float,
            ]
        ] = column


class CreateUnitConverter:
    """Namespace for methods creating unit converters"""

    @staticmethod
    # pylint: disable=invalid-name
    def create_converter_to_SI(
        uscale: float,
    ) -> Callable[[float,], float]:
        """Creates callable that converts a quantity from its measurement units to SI units.

        Example:
        quantity = 100 bar
        [quantity](METRIC) = bar
        [quantity](SI) = Pa
        uscale = METRIC.pressure() = 100 000 Pa/bar
        returns 100 bar * 100 000 Pa/bar = 10 000 000 Pa

        Args:
            uscale: Unit scale of the measurement unit system
                    (e.g. 100 000 Pa for unit system using bar as pressure unit)

        Returns:
            Callable that converts a quantity from its measurement unit system (given by uscale)
            to SI units.
        """
        return lambda quantity: Unit.Convert.from_(quantity, uscale)

    @staticmethod
    def rs_scale(unit_system: EclUnits.UnitSystem) -> float:
        # Rs = [sVolume(Gas) / sVolume(Liquid)]
        return (
            unit_system.surface_volume_gas().value
            / unit_system.surface_volume_liquid().value
        )

    @staticmethod
    def rv_scale(unit_system: EclUnits.UnitSystem) -> float:
        # Rv = [sVolume(Liq) / sVolume(Gas)]
        return (
            unit_system.surface_volume_liquid().value
            / unit_system.surface_volume_gas().value
        )

    @staticmethod
    def fvf_scale(unit_system: EclUnits.UnitSystem) -> float:
        # B = [rVolume / sVolume(Liquid)]
        return (
            unit_system.reservoir_volume().value
            / unit_system.surface_volume_liquid().value
        )

    @staticmethod
    def fvf_gas_scale(unit_system: EclUnits.UnitSystem) -> float:
        # B = [rVolume / sVolume(Gas)]
        return (
            unit_system.reservoir_volume().value
            / unit_system.surface_volume_gas().value
        )

    class ToSI:
        @staticmethod
        def fvf(
            unit_system: EclUnits.UnitSystem,
        ) -> Callable[[float,], float]:
            return CreateUnitConverter.create_converter_to_SI(
                CreateUnitConverter.fvf_scale(unit_system)
            )

        @staticmethod
        def density(
            unit_system: EclUnits.UnitSystem,
        ) -> Callable[[float,], float]:
            return CreateUnitConverter.create_converter_to_SI(
                unit_system.density().value
            )

        @staticmethod
        def pressure(
            unit_system: EclUnits.UnitSystem,
        ) -> Callable[[float,], float]:
            return CreateUnitConverter.create_converter_to_SI(
                unit_system.pressure().value
            )

        @staticmethod
        def compressibility(
            unit_system: EclUnits.UnitSystem,
        ) -> Callable[[float,], float]:
            return CreateUnitConverter.create_converter_to_SI(
                1.0 / unit_system.pressure().value
            )

        @staticmethod
        def viscosity(
            unit_system: EclUnits.UnitSystem,
        ) -> Callable[[float,], float]:
            return CreateUnitConverter.create_converter_to_SI(
                unit_system.viscosity().value
            )

        @staticmethod
        def dissolved_gas_oil_ratio(
            unit_system: EclUnits.UnitSystem,
        ) -> Callable[[float,], float]:
            return CreateUnitConverter.create_converter_to_SI(
                CreateUnitConverter.rs_scale(unit_system)
            )

        @staticmethod
        def vaporised_oil_gas_ratio(
            unit_system: EclUnits.UnitSystem,
        ) -> Callable[[float,], float]:
            return CreateUnitConverter.create_converter_to_SI(
                CreateUnitConverter.rv_scale(unit_system)
            )

        @staticmethod
        def recip_fvf(
            unit_system: EclUnits.UnitSystem,
        ) -> Callable[[float,], float]:
            return CreateUnitConverter.create_converter_to_SI(
                1.0 / CreateUnitConverter.fvf_scale(unit_system)
            )

        @staticmethod
        def recip_fvf_deriv_press(
            unit_system: EclUnits.UnitSystem,
        ) -> Callable[[float,], float]:
            # d(1/B)/dp
            b_scale = CreateUnitConverter.fvf_scale(unit_system)
            p_scale = unit_system.pressure().value

            return CreateUnitConverter.create_converter_to_SI(1.0 / (b_scale * p_scale))

        @staticmethod
        def recip_fvf_deriv_vap_oil(
            unit_system: EclUnits.UnitSystem,
        ) -> Callable[[float,], float]:
            # d(1/B)/dRv
            b_scale = CreateUnitConverter.fvf_scale(unit_system)
            rv_scale = CreateUnitConverter.rv_scale(unit_system)

            return CreateUnitConverter.create_converter_to_SI(
                1.0 / (b_scale * rv_scale)
            )

        @staticmethod
        def recip_fvf_visc(
            unit_system: EclUnits.UnitSystem,
        ) -> Callable[[float,], float]:
            b_scale = CreateUnitConverter.fvf_scale(unit_system)
            visc_scale = unit_system.viscosity().value

            return CreateUnitConverter.create_converter_to_SI(
                1.0 / (b_scale * visc_scale)
            )

        @staticmethod
        def recip_fvf_visc_deriv_press(
            unit_system: EclUnits.UnitSystem,
        ) -> Callable[[float,], float]:
            # d(1/(B*mu))/dp
            b_scale = CreateUnitConverter.fvf_scale(unit_system)
            p_scale = unit_system.pressure().value
            visc_scale = unit_system.viscosity().value

            return CreateUnitConverter.create_converter_to_SI(
                1.0 / (b_scale * visc_scale * p_scale)
            )

        @staticmethod
        def recip_fvf_visc_deriv_vap_oil(
            unit_system: EclUnits.UnitSystem,
        ) -> Callable[[float,], float]:
            # d(1/(B*mu))/dRv
            b_scale = CreateUnitConverter.fvf_scale(unit_system)
            visc_scale = unit_system.viscosity().value
            rv_scale = CreateUnitConverter.rv_scale(unit_system)

            return CreateUnitConverter.create_converter_to_SI(
                1.0 / (b_scale * visc_scale * rv_scale)
            )

        @staticmethod
        def recip_fvf_gas(
            unit_system: EclUnits.UnitSystem,
        ) -> Callable[[float,], float]:
            return CreateUnitConverter.create_converter_to_SI(
                1.0 / CreateUnitConverter.fvf_gas_scale(unit_system)
            )

        @staticmethod
        def recip_fvf_gas_deriv_press(
            unit_system: EclUnits.UnitSystem,
        ) -> Callable[[float,], float]:
            # d(1/B)/dp
            b_scale = CreateUnitConverter.fvf_gas_scale(unit_system)
            p_scale = unit_system.pressure().value

            return CreateUnitConverter.create_converter_to_SI(1.0 / (b_scale * p_scale))

        @staticmethod
        def recip_fvf_gas_deriv_vap_oil(
            unit_system: EclUnits.UnitSystem,
        ) -> Callable[[float,], float]:
            # d(1/B)/dRv
            b_scale = CreateUnitConverter.fvf_gas_scale(unit_system)
            rv_scale = CreateUnitConverter.rv_scale(unit_system)

            return CreateUnitConverter.create_converter_to_SI(
                1.0 / (b_scale * rv_scale)
            )

        @staticmethod
        def recip_fvf_gas_visc(
            unit_system: EclUnits.UnitSystem,
        ) -> Callable[[float,], float]:
            b_scale = CreateUnitConverter.fvf_gas_scale(unit_system)
            visc_scale = unit_system.viscosity().value

            return CreateUnitConverter.create_converter_to_SI(
                1.0 / (b_scale * visc_scale)
            )

        @staticmethod
        def recip_fvf_gas_visc_deriv_press(
            unit_system: EclUnits.UnitSystem,
        ) -> Callable[[float,], float]:
            # d(1/(B*mu))/dp
            b_scale = CreateUnitConverter.fvf_gas_scale(unit_system)
            p_scale = unit_system.pressure().value
            visc_scale = unit_system.viscosity().value

            return CreateUnitConverter.create_converter_to_SI(
                1.0 / (b_scale * visc_scale * p_scale)
            )

        @staticmethod
        def recip_fvf_gas_visc_deriv_vap_oil(
            unit_system: EclUnits.UnitSystem,
        ) -> Callable[[float,], float]:
            # d(1/(B*mu))/dRv
            b_scale = CreateUnitConverter.fvf_gas_scale(unit_system)
            visc_scale = unit_system.viscosity().value
            rv_scale = CreateUnitConverter.rv_scale(unit_system)

            return CreateUnitConverter.create_converter_to_SI(
                1.0 / (b_scale * visc_scale * rv_scale)
            )



import copy
import warnings
from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple

import webviz_subsurface_components as wsc
from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.deprecation_decorators import deprecated_plugin_arguments
from webviz_config.utils import StrEnum

from webviz_subsurface._abbreviations.reservoir_simulation import (
    historical_vector,
    simulation_vector_description,
)
from webviz_subsurface._providers import Frequency
from webviz_subsurface._utils.ensemble_summary_provider_set_factory import (
    create_lazy_ensemble_summary_provider_set_from_paths,
    create_presampled_ensemble_summary_provider_set_from_paths,
)
from webviz_subsurface._utils.simulation_timeseries import (
    check_and_format_observations,
    set_simulation_line_shape_fallback,
)
from webviz_subsurface._utils.user_defined_vector_definitions import (
    create_user_defined_vector_descriptions_from_config,
)
from webviz_subsurface._utils.vector_calculator import (
    add_expressions_to_vector_selector_data,
    expressions_from_config,
    get_vector_definitions_from_expressions,
    validate_predefined_expression,
)
from webviz_subsurface._utils.vector_selector import (
    add_vector_to_vector_selector_data,
    is_vector_name_in_vector_selector_data,
)
from webviz_subsurface._utils.webvizstore_functions import get_path

from ._views._subplot_view import SubplotView
from ._views._subplot_view._settings import (
    EnsemblesSettings,
    FilterRealizationSettings,
    GroupBySettings,
    ResamplingFrequencySettings,
    TimeSeriesSettings,
    VisualizationSettings,
)
from ._views._subplot_view._types import VisualizationOptions
from ._views._subplot_view._utils.from_timeseries_cumulatives import (
    create_per_day_vector_name,
    create_per_interval_vector_name,
)
from ._views._subplot_view._view_elements._subplot_graph import SubplotGraph


def check_deprecation_argument(options: Optional[dict]) -> Optional[Tuple[str, str]]:
    if options is None:
        return None
    if any(elm in options for elm in ["vector1", "vector2", "vector3"]):
        return (
            'The usage of "vector1", "vector2" and "vector3" as user input options are deprecated. '
            'Please replace options with list named "vectors"',
            'The usage of "vector1", "vector2" and "vector3" as user input in options for '
            "initially selected vectors are deprecated. Please replace user input options with "
            'list named "vectors", where each element represent the corresponding initially '
            "selected vector.",
        )
    return None


# pylint: disable=too-many-instance-attributes
class SimulationTimeSeries(WebvizPluginABC):
    class Ids(StrEnum):
        SUBPLOT_VIEW = "subplot-view"

    # pylint: disable=too-many-arguments,too-many-branches,too-many-locals,too-many-statements
    @deprecated_plugin_arguments(check_deprecation_argument)
    def __init__(
        self,
        webviz_settings: WebvizSettings,
        ensembles: Optional[list] = None,
        rel_file_pattern: str = "share/results/unsmry/*.arrow",
        perform_presampling: bool = False,
        obsfile: Path = None,
        options: dict = None,
        sampling: str = Frequency.MONTHLY.value,
        predefined_expressions: str = None,
        user_defined_vector_definitions: str = None,
        line_shape_fallback: str = "linear",
    ) -> None:
        super().__init__(stretch=True)

        self._webviz_settings = webviz_settings
        self._obsfile = obsfile

        # Retrieve user defined vector descriptions from configuration and validate
        self._user_defined_vector_descriptions_path = (
            None
            if user_defined_vector_definitions is None
            else webviz_settings.shared_settings["user_defined_vector_definitions"][
                user_defined_vector_definitions
            ]
        )
        self._user_defined_vector_definitions: Dict[
            str, wsc.VectorDefinition
        ] = create_user_defined_vector_descriptions_from_config(
            get_path(self._user_defined_vector_descriptions_path)
            if self._user_defined_vector_descriptions_path
            else None
        )
        self._custom_vector_definitions = copy.deepcopy(
            self._user_defined_vector_definitions
        )

        self._line_shape_fallback = set_simulation_line_shape_fallback(
            line_shape_fallback
        )

        if Frequency.from_string_value(sampling) is None:
            raise ValueError(
                'Sampling frequency conversion is "None", i.e. Raw sampling, and '
                "is not supported by plugin yet!"
            )
        self._sampling = Frequency(sampling)
        self._presampled_frequency = None

        # TODO: Update functionality when allowing raw data and csv file input
        # NOTE: If csv is implemented-> handle/disable statistics, PER_INTVL_, PER_DAY_, delta
        # ensemble, etc.
        if ensembles is not None:
            ensemble_paths: Dict[str, Path] = {
                ensemble_name: webviz_settings.shared_settings["scratch_ensembles"][
                    ensemble_name
                ]
                for ensemble_name in ensembles
            }
            if perform_presampling:
                self._presampled_frequency = self._sampling
                self._input_provider_set = (
                    create_presampled_ensemble_summary_provider_set_from_paths(
                        ensemble_paths, rel_file_pattern, self._presampled_frequency
                    )
                )
            else:
                self._input_provider_set = (
                    create_lazy_ensemble_summary_provider_set_from_paths(
                        ensemble_paths, rel_file_pattern
                    )
                )
        else:
            raise ValueError('Incorrect argument, must provide "ensembles"')

        if not self._input_provider_set:
            raise ValueError(
                "Initial provider set is undefined, and ensemble summary providers"
                " are not instantiated for plugin"
            )

        self._theme = webviz_settings.theme

        self._observations = {}
        if self._obsfile:
            self._observations = check_and_format_observations(get_path(self._obsfile))

        # NOTE: Initially keep set of all vector names - can make dynamic if wanted?
        vector_names = self._input_provider_set.all_vector_names()
        non_historical_vector_names = [
            vector
            for vector in vector_names
            if historical_vector(vector, None, False) not in vector_names
        ]

        # NOTE: Initially: With set of vector names, the vector selector data is static
        # Can be made dynamic based on selected ensembles - i.e. vectors present among
        # selected providers?
        self._vector_selector_base_data: list = []
        self._vector_calculator_data: list = []
        for vector in non_historical_vector_names:
            add_vector_to_vector_selector_data(
                self._vector_selector_base_data,
                vector,
            )

            # Only vectors from providers are provided to vector calculator
            add_vector_to_vector_selector_data(
                self._vector_calculator_data,
                vector,
            )

            metadata = (
                self._input_provider_set.vector_metadata(vector)
                if self._input_provider_set
                else None
            )
            if metadata and metadata.is_total:
                # Get the likely name for equivalent rate vector and make dropdown options.
                # Requires that the time_index was either defined or possible to infer.
                per_day_vec = create_per_day_vector_name(vector)
                per_intvl_vec = create_per_interval_vector_name(vector)

                add_vector_to_vector_selector_data(
                    self._vector_selector_base_data,
                    per_day_vec,
                )
                add_vector_to_vector_selector_data(
                    self._vector_selector_base_data,
                    per_intvl_vec,
                )

                # Add vector base to custom vector definition if not existing
                vector_base = vector.split(":")[0]
                _definition = wsc.VectorDefinitions.get(vector_base, None)
                _type = _definition["type"] if _definition else "others"

                per_day_vec_base = per_day_vec.split(":")[0]
                per_intvl_vec_base = per_intvl_vec.split(":")[0]
                if per_day_vec_base not in self._custom_vector_definitions:
                    self._custom_vector_definitions[
                        per_day_vec_base
                    ] = wsc.VectorDefinition(
                        type=_type,
                        description=simulation_vector_description(
                            per_day_vec_base, self._user_defined_vector_definitions
                        ),
                    )
                if per_intvl_vec_base not in self._custom_vector_definitions:
                    self._custom_vector_definitions[
                        per_intvl_vec_base
                    ] = wsc.VectorDefinition(
                        type=_type,
                        description=simulation_vector_description(
                            per_intvl_vec_base, self._user_defined_vector_definitions
                        ),
                    )

        # Retrieve predefined expressions from configuration and validate
        self._predefined_expressions_path = (
            None
            if predefined_expressions is None
            else webviz_settings.shared_settings["predefined_expressions"][
                predefined_expressions
            ]
        )
        self._predefined_expressions = expressions_from_config(
            get_path(self._predefined_expressions_path)
            if self._predefined_expressions_path
            else None
        )
        for expression in self._predefined_expressions:
            valid, message = validate_predefined_expression(
                expression, self._vector_selector_base_data
            )
            if not valid:
                warnings.warn(message)
            expression["isValid"] = valid

        # Add expressions to custom vector definitions
        self._custom_vector_definitions_base = copy.deepcopy(
            self._custom_vector_definitions
        )
        _custom_vector_definitions_from_expressions = (
            get_vector_definitions_from_expressions(self._predefined_expressions)
        )
        for key, value in _custom_vector_definitions_from_expressions.items():
            if key not in self._custom_vector_definitions:
                self._custom_vector_definitions[key] = value

        # Create initial vector selector data with predefined expressions
        self._initial_vector_selector_data = copy.deepcopy(
            self._vector_selector_base_data
        )
        add_expressions_to_vector_selector_data(
            self._initial_vector_selector_data, self._predefined_expressions
        )

        plot_options = options if options else {}
        self._initial_visualization_selection = VisualizationOptions(
            plot_options.get("visualization", "statistics")
        )

        # Initial selected vectors - NB: {vector1, vector2, vector3} is deprecated!
        initial_vectors: List[str] = plot_options.get("vectors", [])

        # TODO: Remove when deprecated code is not utilized anymore
        if "vectors" in plot_options and any(
            elm in plot_options for elm in ["vector1", "vector2", "vector3"]
        ):
            warnings.warn(
                'Providing new user input option "vectors" and deprecated user input options '
                '"vector1", "vector2" and "vector3" simultaneously. Initially selected vectors '
                'for plugin are set equal to new user input option "vectors".'
            )
        if not initial_vectors:
            initial_vectors = [
                plot_options[elm]
                for elm in ["vector1", "vector2", "vector3"]
                if elm in plot_options
            ][:3]

        # Check if initially selected vectors exist in data, raise ValueError if not
        missing_vectors = [
            elm
            for elm in initial_vectors
            if not is_vector_name_in_vector_selector_data(
                elm, self._initial_vector_selector_data
            )
        ]
        if missing_vectors:
            raise ValueError(
                f"Cannot find: {', '.join(missing_vectors)} to plot initially in "
                "SimulationTimeSeries. Check that the vector(s) exist in your data."
            )

        _max_num_initial_vectors = 5
        if len(initial_vectors) > _max_num_initial_vectors:
            warnings.warn(
                f'User input option "vectors" contains more than {_max_num_initial_vectors} '
                f"vectors. Only the first {_max_num_initial_vectors} listed vectors are kept "
                "for initially selected vectors - the remaining are neglected."
            )
        self._initial_vectors = initial_vectors[:_max_num_initial_vectors]

        self.add_view(
            SubplotView(
                custom_vector_definitions=self._custom_vector_definitions,
                custom_vector_definitions_base=self._custom_vector_definitions_base,
                disable_resampling_dropdown=self._presampled_frequency is not None,
                initial_selected_vectors=self._initial_vectors,
                initial_vector_selector_data=self._initial_vector_selector_data,
                initial_visualization=self._initial_visualization_selection,
                input_provider_set=self._input_provider_set,
                predefined_expressions=self._predefined_expressions,
                selected_resampling_frequency=self._sampling,
                vector_calculator_data=self._vector_calculator_data,
                vector_selector_base_data=self._vector_selector_base_data,
                theme=self._theme,
                user_defined_vector_definitions=self._user_defined_vector_definitions,
                observations=self._observations,
                line_shape_fallback=self._line_shape_fallback,
            ),
            SimulationTimeSeries.Ids.SUBPLOT_VIEW,
        )

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        functions: List[Tuple[Callable, list]] = []
        if self._obsfile:
            functions.append((get_path, [{"path": self._obsfile}]))
        if self._predefined_expressions_path:
            functions.append((get_path, [{"path": self._predefined_expressions_path}]))
        if self._user_defined_vector_descriptions_path:
            functions.append(
                (get_path, [{"path": self._user_defined_vector_descriptions_path}])
            )
        return functions

    @property
    def tour_steps(self) -> List[dict]:
        return [
            {
                "id": self.view(SimulationTimeSeries.Ids.SUBPLOT_VIEW)
                .view_element(SubplotView.Ids.SUBPLOT)
                .component_unique_id(SubplotGraph.Ids.GRAPH),
                "content": (
                    "Visualization of selected time series. "
                    "Different options can be set in the menu to the left."
                ),
            },
            {
                "id": self.view(SimulationTimeSeries.Ids.SUBPLOT_VIEW)
                .settings_group(SubplotView.Ids.GROUP_BY_SETTINGS)
                .component_unique_id(
                    GroupBySettings.Ids.SUBPLOT_OWNER_OPTIONS_RADIO_ITEMS
                ),
                "content": (
                    "Setting to group visualization data according to selection. "
                    "Subplot per selected time series or per selected ensemble."
                ),
            },
            {
                "id": self.view(SimulationTimeSeries.Ids.SUBPLOT_VIEW)
                .settings_group(SubplotView.Ids.RESAMPLING_FREQUENCY_SETTINGS)
                .component_unique_id(
                    ResamplingFrequencySettings.Ids.RESAMPLING_FREQUENCY_DROPDOWN
                ),
                "content": (
                    "Select resampling frequency for the time series data. "
                    "With presampled data, the dropdown is disabled and the pre-sampling "
                    "frequency shown."
                ),
            },
            {
                "id": self.view(SimulationTimeSeries.Ids.SUBPLOT_VIEW)
                .settings_group(SubplotView.Ids.RESAMPLING_FREQUENCY_SETTINGS)
                .component_unique_id(
                    ResamplingFrequencySettings.Ids.RELATIVE_DATE_DROPDOWN
                ),
                "content": (
                    "Select date to create relative time series data. "
                    "The sample value on selected date is subtracted from each sample in "
                    "the time series to provide data relative to the selected date. "
                ),
            },
            {
                "id": self.view(SimulationTimeSeries.Ids.SUBPLOT_VIEW)
                .settings_group(SubplotView.Ids.ENSEMBLE_SETTINGS)
                .component_unique_id(EnsemblesSettings.Ids.ENSEMBLES_DROPDOWN),
                "content": (
                    "Display time series from one or several ensembles. "
                    "Ensembles will be overlain in subplot or represented per subplot, "
                    'based on selection in "Group By".'
                ),
            },
            {
                "id": self.view(SimulationTimeSeries.Ids.SUBPLOT_VIEW)
                .settings_group(SubplotView.Ids.ENSEMBLE_SETTINGS)
                .component_unique_id(EnsemblesSettings.Ids.DELTA_ENSEMBLE),
                "content": (
                    "Create delta ensembles (A-B). "
                    "Define delta between two ensembles and make available among "
                    "selectable ensembles."
                ),
            },
            {
                "id": self.view(SimulationTimeSeries.Ids.SUBPLOT_VIEW)
                .settings_group(SubplotView.Ids.TIME_SERIES_SETTINGS)
                .component_unique_id(TimeSeriesSettings.Ids.VECTOR_SELECTOR),
                "content": (
                    "Display different time series. Data for each time series will be visualized"
                    ' in subplots based on selection in "Group By".'
                    "Vectors prefixed with PER_DAY_ and PER_INTVL_ are calculated in the fly "
                    "from cumulative vectors, providing average rates and interval cumulatives "
                    "over a time interval from the selected resampling frequency. Vectors "
                    "categorized as calculated are created using the Vector Calculator below."
                ),
            },
            {
                "id": self.view(SimulationTimeSeries.Ids.SUBPLOT_VIEW)
                .settings_group(SubplotView.Ids.TIME_SERIES_SETTINGS)
                .component_unique_id(
                    TimeSeriesSettings.Ids.VECTOR_CALCULATOR_OPEN_BUTTON
                ),
                "content": (
                    "Create mathematical expressions with provided vector time series. "
                    "Parsing of the mathematical expression is handled and will give feedback "
                    "when entering invalid expressions. "
                    "The expressions are calculated on the fly and can be selected among the time "
                    "series to be shown in the visualization."
                ),
            },
            {
                "id": self.view(SimulationTimeSeries.Ids.SUBPLOT_VIEW)
                .settings_group(SubplotView.Ids.VISUALIZATION_SETTINGS)
                .component_unique_id(
                    VisualizationSettings.Ids.VISUALIZATION_RADIO_ITEMS
                ),
                "content": (
                    "Choose between different visualizations. 1. Show time series as "
                    "individual lines per realization. 2. Show statistical lines per "
                    "ensemble. 3. Show statistical fanchart per ensemble. 4. Show "
                    "statistical lines per ensemble and individual lines per realization "
                    "simultaneously"
                ),
            },
            {
                "id": self.view(SimulationTimeSeries.Ids.SUBPLOT_VIEW)
                .settings_group(SubplotView.Ids.VISUALIZATION_SETTINGS)
                .component_unique_id(VisualizationSettings.Ids.PLOT_OPTIONS),
                "content": (
                    "Various plot options: Whether to include history trace or vector observations "
                    "and which statistics to show if statistical lines or fanchart is chosen as "
                    "visualization."
                ),
            },
            {
                "id": self.view(SimulationTimeSeries.Ids.SUBPLOT_VIEW)
                .settings_group(SubplotView.Ids.FILTER_REALIZATION_SETTINGS)
                .component_unique_id(
                    FilterRealizationSettings.Ids.STATISTICS_FROM_RADIO_ITEMS
                ),
                "content": (
                    "Select whether to calculate statistics from all realizations, or to calculate "
                    "statistics from the selected subset of realizations "
                ),
            },
            {
                "id": self.view(SimulationTimeSeries.Ids.SUBPLOT_VIEW)
                .settings_group(SubplotView.Ids.FILTER_REALIZATION_SETTINGS)
                .component_unique_id(
                    FilterRealizationSettings.Ids.REALIZATIONS_FILTER_SELECTOR
                ),
                "content": (
                    "Filter realizations. Select realization numbers to include in visualization, "
                    "and utilize in statistics calculation when calculating from selected subset."
                ),
            },
        ]



from typing import Any, Dict, List, Optional, Tuple

import plotly.graph_objects as go
from dash import Dash, Input, Output, State, callback, html
from dash.exceptions import PreventUpdate
from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.utils import StrEnum

from webviz_subsurface._providers import FaultPolygonsServer, SurfaceImageServer
from webviz_subsurface.plugins._co2_leakage._utilities.callbacks import (
    SurfaceData,
    create_map_layers,
    derive_surface_address,
    get_plume_polygon,
    property_origin,
    readable_name,
)
from webviz_subsurface.plugins._co2_leakage._utilities.co2volume import (
    generate_co2_time_containment_figure,
    generate_co2_volume_figure,
)
from webviz_subsurface.plugins._co2_leakage._utilities.fault_polygons import (
    FaultPolygonsHandler,
)
from webviz_subsurface.plugins._co2_leakage._utilities.generic import MapAttribute
from webviz_subsurface.plugins._co2_leakage._utilities.initialization import (
    init_co2_containment_table_providers,
    init_map_attribute_names,
    init_surface_providers,
    init_well_pick_provider,
)
from webviz_subsurface.plugins._co2_leakage.views.mainview.mainview import (
    MainView,
    MapViewElement,
)
from webviz_subsurface.plugins._co2_leakage.views.mainview.settings import ViewSettings

from . import _error
from ._utilities.color_tables import co2leakage_color_tables


class CO2Leakage(WebvizPluginABC):
    """
    Plugin for analyzing CO2 leakage potential across multiple realizations in an FMU
    ensemble

    * **`ensembles`:** Which ensembles in `shared_settings` to visualize.
    * **`boundary_file`:** Path to a polygon representing the containment area
    * **`well_pick_file`:** Path to a file containing well picks
    * **`co2_containment_relpath`:** Path to a table of co2 containment data (amount of
        CO2 outside/inside a boundary). Relative to each realization.
    * **`fault_polygon_attribute`:** Polygons with this attribute are used as fault
        polygons
    * **`map_attribute_names`:** Dictionary for overriding the default mapping between
        attributes visualized by the plugin and attributes names used by
        EnsembleSurfaceProvider
    * **`initial_surface`:** Name of the surface/formation to show when the plugin is
        launched. If no name is provided, the first alphabetical surface is shown.
    * **`map_surface_names_to_well_pick_names`:** Optional mapping between surface map
        names and surface names used in the well pick file
    * **`map_surface_names_to_fault_polygons`:** Optional mapping between surface map
        names and surface names used by the fault polygons
    """

    class Ids(StrEnum):
        MAIN_VIEW = "main-view"
        MAIN_SETTINGS = "main-settings"

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        app: Dash,
        webviz_settings: WebvizSettings,
        ensembles: List[str],
        boundary_file: Optional[str] = None,
        well_pick_file: Optional[str] = None,
        co2_containment_relpath: str = "share/results/tables/co2_volumes.csv",
        fault_polygon_attribute: str = "dl_extracted_faultlines",
        initial_surface: Optional[str] = None,
        map_attribute_names: Optional[Dict[str, str]] = None,
        map_surface_names_to_well_pick_names: Optional[Dict[str, str]] = None,
        map_surface_names_to_fault_polygons: Optional[Dict[str, str]] = None,
    ):
        super().__init__()
        self._error_message = ""

        self._boundary_file = boundary_file
        try:
            self._ensemble_paths = {
                ensemble_name: webviz_settings.shared_settings["scratch_ensembles"][
                    ensemble_name
                ]
                for ensemble_name in ensembles
            }
            self._surface_server = SurfaceImageServer.instance(app)
            self._polygons_server = FaultPolygonsServer.instance(app)

            self._map_attribute_names = init_map_attribute_names(map_attribute_names)
            # Surfaces
            self._ensemble_surface_providers = init_surface_providers(
                webviz_settings, ensembles
            )
            # Polygons
            self._fault_polygon_handlers = {
                ens: FaultPolygonsHandler(
                    self._polygons_server,
                    self._ensemble_paths[ens],
                    map_surface_names_to_fault_polygons or {},
                    fault_polygon_attribute,
                )
                for ens in ensembles
            }
            # CO2 containment
            self._co2_table_providers = init_co2_containment_table_providers(
                self._ensemble_paths,
                co2_containment_relpath,
            )
            # Well picks
            self._well_pick_provider = init_well_pick_provider(
                well_pick_file,
                map_surface_names_to_well_pick_names,
            )
        except Exception as err:
            self._error_message = f"Plugin initialization failed: {err}"
            raise

        color_tables = co2leakage_color_tables()
        self.add_shared_settings_group(
            ViewSettings(
                self._ensemble_paths,
                self._ensemble_surface_providers,
                initial_surface,
                self._map_attribute_names,
                [c["name"] for c in color_tables],  # type: ignore
            ),
            self.Ids.MAIN_SETTINGS,
        )
        self.add_view(MainView(color_tables), self.Ids.MAIN_VIEW)

    @property
    def layout(self) -> html.Div:
        return _error.error(self._error_message)

    def _view_component(self, component_id: str) -> str:
        return (
            self.view(self.Ids.MAIN_VIEW)
            .view_element(MainView.Ids.MAIN_ELEMENT)
            .component_unique_id(component_id)
            .to_string()
        )

    def _settings_component(self, component_id: str) -> str:
        return (
            self.shared_settings_group(self.Ids.MAIN_SETTINGS)
            .component_unique_id(component_id)
            .to_string()
        )

    def _ensemble_dates(self, ens: str) -> List[str]:
        surface_provider = self._ensemble_surface_providers[ens]
        att_name = self._map_attribute_names[MapAttribute.MAX_SGAS]
        dates = surface_provider.surface_dates_for_attribute(att_name)
        if dates is None:
            raise ValueError(f"Failed to fetch dates for attribute '{att_name}'")
        return dates

    def _set_callbacks(self) -> None:
        @callback(
            Output(self._view_component(MapViewElement.Ids.BAR_PLOT), "figure"),
            Output(self._view_component(MapViewElement.Ids.TIME_PLOT), "figure"),
            Input(self._settings_component(ViewSettings.Ids.ENSEMBLE), "value"),
        )
        def update_graphs(ensemble: str) -> Tuple[go.Figure, go.Figure]:
            fig_args = (
                self._co2_table_providers[ensemble],
                self._co2_table_providers[ensemble].realizations(),
            )
            fig0 = generate_co2_volume_figure(*fig_args)
            fig1 = generate_co2_time_containment_figure(*fig_args)
            return fig0, fig1

        @callback(
            Output(self._view_component(MapViewElement.Ids.DATE_SLIDER), "marks"),
            Output(self._view_component(MapViewElement.Ids.DATE_SLIDER), "value"),
            Input(self._settings_component(ViewSettings.Ids.ENSEMBLE), "value"),
        )
        def set_dates(ensemble: str) -> Tuple[Dict[int, Dict[str, Any]], Optional[int]]:
            if ensemble is None:
                return {}, None
            # Dates
            date_list = self._ensemble_dates(ensemble)
            dates = {
                i: {
                    "label": f"{d[:4]}",
                    "style": {"writingMode": "vertical-rl"},
                }
                for i, d in enumerate(date_list)
            }
            initial_date = max(dates.keys())
            return dates, initial_date

        @callback(
            Output(self._view_component(MapViewElement.Ids.DATE_WRAPPER), "style"),
            Input(self._settings_component(ViewSettings.Ids.PROPERTY), "value"),
        )
        def toggle_date_slider(attribute: str) -> Dict[str, str]:
            if MapAttribute(attribute) == MapAttribute.MIGRATION_TIME:
                return {"display": "none"}
            return {}

        # Cannot avoid many arguments and/or locals since all layers of the DeckGL map
        # needs to be updated simultaneously
        # pylint: disable=too-many-arguments,too-many-locals
        @callback(
            Output(self._view_component(MapViewElement.Ids.DECKGL_MAP), "layers"),
            Input(self._settings_component(ViewSettings.Ids.PROPERTY), "value"),
            Input(self._view_component(MapViewElement.Ids.DATE_SLIDER), "value"),
            Input(self._settings_component(ViewSettings.Ids.FORMATION), "value"),
            Input(self._settings_component(ViewSettings.Ids.REALIZATION), "value"),
            Input(self._settings_component(ViewSettings.Ids.STATISTIC), "value"),
            Input(self._settings_component(ViewSettings.Ids.COLOR_SCALE), "value"),
            Input(self._settings_component(ViewSettings.Ids.CM_MIN_AUTO), "value"),
            Input(self._settings_component(ViewSettings.Ids.CM_MIN), "value"),
            Input(self._settings_component(ViewSettings.Ids.CM_MAX_AUTO), "value"),
            Input(self._settings_component(ViewSettings.Ids.CM_MAX), "value"),
            Input(self._settings_component(ViewSettings.Ids.PLUME_THRESHOLD), "value"),
            Input(self._settings_component(ViewSettings.Ids.PLUME_SMOOTHING), "value"),
            State(self._settings_component(ViewSettings.Ids.ENSEMBLE), "value"),
        )
        def update_map_attribute(
            attribute: MapAttribute,
            date: int,
            formation: str,
            realization: List[int],
            statistic: str,
            color_map_name: str,
            cm_min_auto: List[str],
            cm_min_val: Optional[float],
            cm_max_auto: List[str],
            cm_max_val: Optional[float],
            plume_threshold: Optional[float],
            plume_smoothing: Optional[float],
            ensemble: str,
        ) -> List[Dict[str, Any]]:
            attribute = MapAttribute(attribute)
            if len(realization) == 0:
                raise PreventUpdate
            if ensemble is None:
                raise PreventUpdate
            datestr = self._ensemble_dates(ensemble)[date]
            # Contour data
            contour_data = None
            if attribute in (MapAttribute.SGAS_PLUME, MapAttribute.AMFG_PLUME):
                contour_data = {
                    "property": property_origin(attribute, self._map_attribute_names),
                    "threshold": plume_threshold,
                    "smoothing": plume_smoothing,
                }
            # Surface
            surf_data = None
            if formation is not None and len(realization) > 0:
                surf_data = SurfaceData.from_server(
                    server=self._surface_server,
                    provider=self._ensemble_surface_providers[ensemble],
                    address=derive_surface_address(
                        formation,
                        attribute,
                        datestr,
                        realization,
                        self._map_attribute_names,
                        statistic,
                        contour_data,
                    ),
                    color_map_range=(
                        cm_min_val if len(cm_min_auto) == 0 else None,
                        cm_max_val if len(cm_max_auto) == 0 else None,
                    ),
                    color_map_name=color_map_name,
                    readable_name_=readable_name(attribute),
                )
            # Plume polygon
            plume_polygon = None
            if contour_data is not None:
                plume_polygon = get_plume_polygon(
                    self._ensemble_surface_providers[ensemble],
                    realization,
                    formation,
                    datestr,
                    contour_data,
                )
            # Create layers and view bounds
            layers = create_map_layers(
                formation=formation,
                surface_data=surf_data,
                fault_polygon_url=(
                    self._fault_polygon_handlers[ensemble].extract_fault_polygon_url(
                        formation,
                        realization,
                    )
                ),
                license_boundary_file=self._boundary_file,
                well_pick_provider=self._well_pick_provider,
                plume_extent_data=plume_polygon,
            )

            return layers



import datetime

import pandas as pd
import pytest
from pandas._testing import assert_frame_equal

from webviz_subsurface._utils.dataframe_utils import make_date_column_datetime_object
from webviz_subsurface.plugins._simulation_time_series._views._subplot_view._types import (
    StatisticsOptions,
)

# pylint: disable=line-too-long
from webviz_subsurface.plugins._simulation_time_series._views._subplot_view._utils.vector_statistics import (
    create_vectors_statistics_df,
)

# *******************************************************************
#####################################################################
#
# CONFIGURE TESTDATA
#
#####################################################################
# *******************************************************************

# fmt: off
INPUT_YEAR_2020_DF = pd.DataFrame(
    columns = ["DATE", "REAL",  "A", "B"],
    data = [
        [datetime.datetime(2020,1,1),  1,      10.0,       50.0    ],
        [datetime.datetime(2020,2,1),  1,      100.0,      500.0   ],
        [datetime.datetime(2020,3,1),  1,      1000.0,     5000.0  ],
        [datetime.datetime(2020,4,1),  1,      10000.0,    50000.0 ],
        [datetime.datetime(2020,5,1),  1,      100000.0,   500000.0],
        [datetime.datetime(2020,1,1),  2,      20.0,       60.0    ],
        [datetime.datetime(2020,2,1),  2,      200.0,      600.0   ],
        [datetime.datetime(2020,3,1),  2,      2000.0,     6000.0  ],
        [datetime.datetime(2020,4,1),  2,      20000.0,    60000.0 ],
        [datetime.datetime(2020,5,1),  2,      200000.0,   600000.0],
        [datetime.datetime(2020,1,1),  4,      30.0,       70.0    ],
        [datetime.datetime(2020,2,1),  4,      300.0,      700.0   ],
        [datetime.datetime(2020,3,1),  4,      3000.0,     7000.0  ],
        [datetime.datetime(2020,4,1),  4,      30000.0,    70000.0 ],
        [datetime.datetime(2020,5,1),  4,      300000.0,   700000.0],
        [datetime.datetime(2020,1,1),  5,      40.0,       80.0    ],
        [datetime.datetime(2020,2,1),  5,      400.0,      800.0   ],
        [datetime.datetime(2020,3,1),  5,      4000.0,     8000.0  ],
        [datetime.datetime(2020,4,1),  5,      40000.0,    80000.0 ],
        [datetime.datetime(2020,5,1),  5,      400000.0,   800000.0]
    ]
)
make_date_column_datetime_object(INPUT_YEAR_2020_DF)

# pylint: disable=line-too-long
# Columns are:
#           ["DATE", "A"                             "B"                          ]
#           [        MEAN, MIN, MAX, P10, P90, P50,  MEAN, MIN, MAX, P10, P90, P50]
# NOTE: P10 is 90 percentil and P90 is 10 percentile for oil standard
EXPECTED_STATISTICS_YEAR_2020_DF = pd.DataFrame(
    columns = pd.MultiIndex.from_tuples([
        ("DATE", ""),
        ("A", StatisticsOptions.MEAN),
        ("A", StatisticsOptions.MIN),
        ("A", StatisticsOptions.MAX),
        ("A", StatisticsOptions.P10),
        ("A", StatisticsOptions.P90),
        ("A", StatisticsOptions.P50),
        ("B", StatisticsOptions.MEAN),
        ("B", StatisticsOptions.MIN),
        ("B", StatisticsOptions.MAX),
        ("B", StatisticsOptions.P10),
        ("B", StatisticsOptions.P90),
        ("B", StatisticsOptions.P50),
    ]),
    data = [
        [datetime.datetime(2020,1,1), 25.0,     10.0,     40.0,     37.0,     13.0,     25.0,     65.0,     50.0,     80.0,     77.0,     53.0,     65.0    ],
        [datetime.datetime(2020,2,1), 250.0,    100.0,    400.0,    370.0,    130.0,    250.0,    650.0,    500.0,    800.0,    770.0,    530.0,    650.0   ],
        [datetime.datetime(2020,3,1), 2500.0,   1000.0,   4000.0,   3700.0,   1300.0,   2500.0,   6500.0,   5000.0,   8000.0,   7700.0,   5300.0,   6500.0  ],
        [datetime.datetime(2020,4,1), 25000.0,  10000.0,  40000.0,  37000.0,  13000.0,  25000.0,  65000.0,  50000.0,  80000.0,  77000.0,  53000.0,  65000.0 ],
        [datetime.datetime(2020,5,1), 250000.0, 100000.0, 400000.0, 370000.0, 130000.0, 250000.0, 650000.0, 500000.0, 800000.0, 770000.0, 530000.0, 650000.0],
    ]
)
make_date_column_datetime_object(EXPECTED_STATISTICS_YEAR_2020_DF)
# fmt: on

# Dates AFTER year 2262!
INPUT_YEAR_2263_DATES = pd.Series(
    [
        datetime.datetime(2263, 1, 1),
        datetime.datetime(2263, 2, 1),
        datetime.datetime(2263, 3, 1),
        datetime.datetime(2263, 4, 1),
        datetime.datetime(2263, 5, 1),
        datetime.datetime(2263, 1, 1),
        datetime.datetime(2263, 2, 1),
        datetime.datetime(2263, 3, 1),
        datetime.datetime(2263, 4, 1),
        datetime.datetime(2263, 5, 1),
        datetime.datetime(2263, 1, 1),
        datetime.datetime(2263, 2, 1),
        datetime.datetime(2263, 3, 1),
        datetime.datetime(2263, 4, 1),
        datetime.datetime(2263, 5, 1),
        datetime.datetime(2263, 1, 1),
        datetime.datetime(2263, 2, 1),
        datetime.datetime(2263, 3, 1),
        datetime.datetime(2263, 4, 1),
        datetime.datetime(2263, 5, 1),
    ]
)
EXPECTED_YEAR_2263_DATES = pd.Series(
    [
        datetime.datetime(2263, 1, 1),
        datetime.datetime(2263, 2, 1),
        datetime.datetime(2263, 3, 1),
        datetime.datetime(2263, 4, 1),
        datetime.datetime(2263, 5, 1),
    ]
)
# NOTE: datetime.datetime after year 2262 is not converted to pd.Timestamp, thus
# no need to make date column datetime object
INPUT_YEAR_2263_DF = INPUT_YEAR_2020_DF.copy()
INPUT_YEAR_2263_DF["DATE"] = INPUT_YEAR_2263_DATES
EXPECTED_STATISTICS_YEAR_2263_DF = EXPECTED_STATISTICS_YEAR_2020_DF.copy()
EXPECTED_STATISTICS_YEAR_2263_DF["DATE"] = EXPECTED_YEAR_2263_DATES

# *******************************************************************
#####################################################################
#
# UNIT TESTS
#
#####################################################################
# *******************************************************************

TEST_VALID_CASES = [
    pytest.param(INPUT_YEAR_2020_DF, EXPECTED_STATISTICS_YEAR_2020_DF),
    pytest.param(INPUT_YEAR_2263_DF, EXPECTED_STATISTICS_YEAR_2263_DF),
]


@pytest.mark.parametrize("input_df, expected_statistics_df", TEST_VALID_CASES)
def test_create_vectors_statistics_df(
    input_df: pd.DataFrame, expected_statistics_df: pd.DataFrame
) -> None:
    statistics_df = create_vectors_statistics_df(input_df)

    assert_frame_equal(statistics_df, expected_statistics_df)


def test_create_vectors_statstics_df_invalid_input() -> None:
    """Test assert check assert_date_column_is_datetime_object() in
    webviz_subsurface._utils.dataframe_utils.py
    """
    # fmt: off
    input_df = pd.DataFrame(
        columns=["DATE", "REAL", "A", "B"],
        data=[
            [pd.Timestamp(2020, 1, 1), 1, 10.0,     50.0    ],
            [pd.Timestamp(2020, 2, 1), 1, 100.0,    500.0   ],
            [pd.Timestamp(2020, 3, 1), 1, 1000.0,   5000.0  ],
            [pd.Timestamp(2020, 4, 1), 1, 10000.0,  50000.0 ],
            [pd.Timestamp(2020, 5, 1), 1, 100000.0, 500000.0],
        ],
    )
    # fmt: on

    with pytest.raises(ValueError) as err:
        create_vectors_statistics_df(input_df)
    assert (
        str(err.value)
        == '"DATE"-column in dataframe is not on datetime.datetime format!'
    )



# pylint: disable=no-name-in-module
from webviz_config.plugins import RelativePermeability


def test_relative_permeability(dash_duo, app, shared_settings) -> None:
    plugin = RelativePermeability(
        app,
        shared_settings["HM_SETTINGS"],
        ensembles=shared_settings["HM_ENSEMBLES"],
        relpermfile="share/results/tables/relperm.csv",
    )
    app.layout = plugin.layout
    dash_duo.start_server(app)
    assert dash_duo.get_logs() == []



import logging
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objs as go
import webviz_core_components as wcc

from ..utils import make_dataframes as makedf

# Color scales
HEATMAP_COLOR = [
    # [0, "gold"],
    [0, "#FF5500"],
    [0.1, "red"],
    [0.3, "darkred"],
    [0.4, "dimgrey"],
    [0.45, "lightgrey"],
    [0.5, "WhiteSmoke"],
    [0.55, "lightgrey"],
    [0.6, "dimgrey"],
    [0.7, "darkblue"],
    [0.9, "blue"],
    [1, "#00AFFF"],
    # [1, "cyan"],
]

# -------------------------------
# pylint: disable = too-many-locals
def prod_misfit_plot(
    df_diff: pd.DataFrame,
    phases: list,
    colorby: str,
    sorting: str = None,
    figheight: int = 450,
    misfit_exponent: float = 1.0,
    normalize: bool = False,
) -> List[wcc.Graph]:
    """Create plot of misfit per realization. One plot per ensemble.
    Misfit is absolute value of |sim - obs|, weighted by obs_error"""

    logging.debug("--- Updating production misfit plot ---")

    if df_diff.empty:
        fig = px.bar(title="No data to plot for current selections.")
        return [wcc.Graph(figure=fig, style={"height": figheight})]

    figures, max_misfit, min_misfit = [], 0, 0

    for ens_name, ensdf in df_diff.groupby("ENSEMBLE"):

        df_misfit, plot_phases, color_phases = _calculate_misfits(
            phases,
            ensdf,
            normalize,
            misfit_exponent,
        )

        # caclulate min-max ranges from first ensemble
        if max_misfit == min_misfit == 0:
            for _, df_date in df_misfit.groupby("DATE"):
                max_misfit = max_misfit + df_date["TOTAL_MISFIT"].max()
                min_misfit = min_misfit + df_date["TOTAL_MISFIT"].min()

        fig = _create_fig_barplot(
            ens_name,
            df_misfit,
            colorby,
            plot_phases,
            color_phases,
            max_misfit,
            min_misfit,
            sorting,
        )

        figures.append(wcc.Graph(figure=fig, style={"height": figheight}))

    return figures


# -------------------------------
def coverage_diffplot(
    df_diff: pd.DataFrame,
    phases: list,
    colorby: str,
    vector_type: str = "well",
    figheight: int = 450,
    boxmode: str = "group",
    boxplot_points: str = "outliers",
) -> List[wcc.Graph]:
    """Create plot of misfit per well. One plot per phase."""

    logging.debug("--- Updating coverage box plot ---")
    figures = []

    if df_diff.empty:
        fig = px.box(title="No data to plot for current selections.")
        return [wcc.Graph(figure=fig, style={"height": figheight})]

    # --- drop columns (realizations) with no data
    ensdf = df_diff.dropna(axis="columns")
    ensdf.DATE = ensdf.DATE.str[:10]

    for phase in phases:

        fig_phase = _create_fig_diffplot(
            ensdf, colorby, vector_type, phase, boxmode, boxplot_points
        )

        figures.append(wcc.Graph(figure=fig_phase, style={"height": figheight}))

    return figures


# -------------------------------
def coverage_crossplot(
    df_smry: pd.DataFrame,
    phases: list,
    colorby: str,
    vector_type: str = "well",
    figheight: int = 450,
    boxplot_points: str = "outliers",
) -> List[wcc.Graph]:
    """Create plot of hist vs simvector. One plot per phase."""

    logging.debug("--- Updating coverage box plot ---")
    figures = []

    prefix = "W" if vector_type == "well" else "G"
    phase_vector, phase_hvector = {}, {}
    phase_vector["Oil"] = prefix + "OPT"
    phase_vector["Water"] = prefix + "WPT"
    phase_vector["Gas"] = prefix + "GPT"
    phase_hvector["Oil"] = prefix + "OPTH"
    phase_hvector["Water"] = prefix + "WPTH"
    phase_hvector["Gas"] = prefix + "GPTH"

    vectorlist = []
    for phase in phases:
        vectorlist.append(phase_vector[phase])
        vectorlist.append(phase_hvector[phase])

    # --- drop columns (realizations) with no data
    # ensdf = df_smry.dropna(axis="columns")

    ensdf = df_smry
    ensdf.DATE = ensdf.DATE.str[:10]
    ensdf["id"] = ensdf.reset_index().index
    ensdf = pd.wide_to_long(
        ensdf,
        vectorlist,
        i="id",
        j="WELL",
        sep=":",
        suffix=r".+",
        # suffix=r"\w+|\d+",
    )
    # use hist avg values as x-values
    ensdf = makedf.get_df_hist_avg(ensdf)

    for phase in phases:

        fig_phase = _create_fig_crossplot(
            ensdf,
            colorby,
            phase,
            boxplot_points,
            phase_vector,
            phase_hvector,
        )

        figures.append(wcc.Graph(figure=fig_phase, style={"height": figheight}))

    return figures


# -------------------------------
def heatmap_plot(
    df_diff_stat: pd.DataFrame,
    phases: list,
    vector_type: str = "well",
    filter_largest: int = 10,
    figheight: int = 450,
    scale_col_range: float = 1.0,
) -> List[wcc.Graph]:
    """Create heatmap of mean misfit per well or group. One plot per phase."""

    logging.debug("--- Updating heatmap ---")
    figures = []

    prefix = "DIFF_W" if vector_type == "well" else "DIFF_G"
    phase_vector = {}
    phase_vector["Oil"] = prefix + "OPT"
    phase_vector["Water"] = prefix + "WPT"
    phase_vector["Gas"] = prefix + "GPT"

    logging.debug(f"Dataframe, diff_stat:\n{df_diff_stat}")

    if df_diff_stat.empty:
        fig = px.bar(title="No data to plot for current selections.")
        return [wcc.Graph(figure=fig, style={"height": figheight})]

    # -------------------------
    for phase in phases:
        df_diff_stat_phase = df_diff_stat[df_diff_stat.VECTOR == phase_vector[phase]]

        # logging.debug(f"Dataframe, diff {phase} phase:\n{df_diff_stat_phase}")

        zmax = scale_col_range * max(
            abs(df_diff_stat_phase.DIFF_MEAN.max()),
            abs(df_diff_stat_phase.DIFF_MEAN.min()),
        )

        for ens_name, df_diff_stat_phase in df_diff_stat_phase.groupby("ENSEMBLE"):

            fig_phase = _create_fig_heatmap(
                ens_name, phase, df_diff_stat_phase, filter_largest, zmax
            )

            figures.append(wcc.Graph(figure=fig_phase, style={"height": figheight}))

    return figures


# -- help functions -------------


def _calculate_misfits(
    phases: List[str],
    ensdf: pd.DataFrame,
    normalize: bool,
    misfit_exponent: float,
) -> Tuple[pd.DataFrame, List[str], Dict[str, str]]:
    """Calculate misfits for total and phases and return as dataframe"""

    # --- drop columns (realizations) with no data
    ensdf = ensdf.dropna(axis="columns")

    all_columns = list(ensdf)  # column names

    df_misfit = ensdf[["ENSEMBLE", "DATE", "REAL"]].copy()
    df_misfit = df_misfit.astype({"REAL": "string"})
    df_misfit["TOTAL_MISFIT"] = 0

    plot_phases = []
    color_phases = {}

    phase_misfit_name = dict(Oil="OIL_MISFIT", Water="WAT_MISFIT", Gas="GAS_MISFIT")
    diff_name = dict(Oil="DIFF_WOPT", Water="DIFF_WWPT", Gas="DIFF_WGPT")
    color = dict(Oil="#2ca02c", Water="#1f77b4", Gas="#d62728")

    for phase in phases:
        phase_columns = [x for x in all_columns if x.startswith(diff_name[phase])]
        df_misfit[phase_misfit_name[phase]] = ensdf[phase_columns].abs().sum(axis=1)
        if normalize:
            df_misfit[phase_misfit_name[phase]] = df_misfit[
                phase_misfit_name[phase]
            ] / len(phase_columns)
        df_misfit[phase_misfit_name[phase]] = df_misfit[phase_misfit_name[phase]] ** (
            1 / misfit_exponent
        )
        df_misfit["TOTAL_MISFIT"] = (
            df_misfit["TOTAL_MISFIT"] + df_misfit[phase_misfit_name[phase]]
        )
        plot_phases.append(phase_misfit_name[phase])
        color_phases[phase_misfit_name[phase]] = color[phase]
    # -------------------------

    return df_misfit, plot_phases, color_phases


# -------------------------------
def _create_fig_barplot(
    ens_name: str,
    df_misfit: pd.DataFrame,
    colorby: str,
    plot_phases: List[str],
    color_phases: Dict[str, str],
    max_misfit: float,
    min_misfit: float,
    sorting: Optional[str],
) -> px.bar:
    """Return misfit bar plot"""

    mean_misfit = df_misfit["TOTAL_MISFIT"].mean()

    color: Any = px.NO_COLOR
    color_discrete_map: Optional[dict] = None
    if colorby == "misfit":
        color = "TOTAL_MISFIT"
    elif colorby == "date":
        color = "DATE"
    elif colorby == "phases":
        color = None
        color_discrete_map = color_phases

    fig = px.bar(
        df_misfit,
        x="REAL",
        y=plot_phases,
        title=ens_name,
        range_y=[0, max_misfit * 1.05],
        color=color,
        color_discrete_map=color_discrete_map,
        range_color=[min_misfit * 0.20, max_misfit * 1.00],
        color_continuous_scale=px.colors.sequential.amp,
    )
    if sorting:
        fig.update_layout(xaxis={"categoryorder": sorting})
    fig.update_xaxes(showticklabels=False)
    fig.update_xaxes(title_text="Realization (hover to see values)")
    fig.update_yaxes(title_text="Cumulative misfit")
    fig.add_hline(mean_misfit)
    fig.add_annotation(average_arrow_annotation(mean_misfit))
    fig.update_layout(margin=dict(l=20, r=20, t=30, b=20))
    # fig.update_layout(coloraxis_colorbar_thickness=20)
    # fig.update(layout_coloraxis_showscale=False)

    return fig


# -------------------------------
def _create_fig_diffplot(
    ensdf: pd.DataFrame,
    colorby: str,
    vector_type: str,
    phase: str,
    boxmode: str,
    boxplot_points: str,
) -> Union[px.box, px.strip]:
    """Return boxplot or strip-plot fig"""

    all_columns = list(ensdf)  # column names

    prefix = "DIFF_W" if vector_type == "well" else "DIFF_G"
    phase_vector = {}
    phase_vector["Oil"] = prefix + "OPT"
    phase_vector["Water"] = prefix + "WPT"
    phase_vector["Gas"] = prefix + "GPT"

    facet_name = "DATE"
    if colorby == "DATE":
        facet_name = "ENSEMBLE"

    phase_columns = [x for x in all_columns if x.startswith(phase_vector[phase])]
    phase_well_labels = [col.split(":")[1] for col in phase_columns]
    text_labels = dict(value=f"{phase} diff (sim-obs)", variable="Well name")

    if boxplot_points == "strip":
        fig_phase = px.strip(
            ensdf,
            y=phase_columns,
            color=colorby,
            facet_col=facet_name,
            facet_col_wrap=2,
            labels=text_labels,
            stripmode=boxmode,
        )
    else:
        fig_phase = px.box(
            ensdf,
            y=phase_columns,
            color=colorby,
            facet_col=facet_name,
            facet_col_wrap=2,
            points=boxplot_points,
            labels=text_labels,
            boxmode=boxmode,
        )
    fig_phase.add_hline(0)
    fig_phase.update_xaxes(ticktext=phase_well_labels, tickvals=phase_columns)

    return fig_phase


# -------------------------------
def _create_fig_crossplot(
    ensdf: pd.DataFrame,
    colorby: str,
    phase: str,
    boxplot_points: str,
    phase_vector: dict,
    phase_hvector: dict,
) -> Union[px.box, px.strip]:
    """Return crossplot style boxplot or strip-plot fig. ensdf must be in long format."""

    facet_name = "DATE"
    if colorby == "DATE":
        facet_name = "ENSEMBLE"

    if boxplot_points == "strip":
        fig_phase = px.strip(
            ensdf,
            x=phase_hvector[phase],
            y=phase_vector[phase],
            color=colorby,
            facet_col=facet_name,
            facet_col_wrap=2,
            hover_name=ensdf.index.get_level_values("WELL"),
        )
    else:
        fig_phase = px.box(
            ensdf,
            x=phase_hvector[phase],
            y=phase_vector[phase],
            color=colorby,
            facet_col=facet_name,
            facet_col_wrap=2,
            points=boxplot_points,
            hover_name=ensdf.index.get_level_values("WELL"),
        )

    # -- add zeroline (diagonal) for oil_vector
    rmin = min(ensdf[phase_hvector[phase]].min(), ensdf[phase_vector[phase]].min())
    rmax = max(ensdf[phase_hvector[phase]].max(), ensdf[phase_vector[phase]].max())
    fig_phase.add_trace(
        go.Scattergl(
            x=[rmin, rmax],
            y=[rmin, rmax],
            mode="lines",
            line_color="rgb(0,100,80)",  # "gray",
            name="zeroline",
            showlegend=True,
        ),
        row="all",
        col="all",
        exclude_empty_subplots=True,
    )

    # -- add 10% off-set for oil_vector
    fig_phase.add_trace(
        go.Scattergl(
            x=[rmin, rmax] + [rmax, rmin],
            y=[rmin * 1.1, rmax * 1.1] + [rmax * 0.9, rmin * 0.9],
            fill="toself",
            fillcolor="rgba(0,100,80,0.2)",
            # mode="lines",
            line_color="rgba(255,255,255,0)",
            name="±10% off-set",
            showlegend=True,
        ),
        row="all",
        col="all",
        exclude_empty_subplots=True,
    )

    # -- add 20% off-set for oil_vector
    fig_phase.add_trace(
        go.Scattergl(
            x=[rmin, rmax] + [rmax, rmin],
            y=[rmin * 1.2, rmax * 1.2] + [rmax * 0.8, rmin * 0.8],
            fill="toself",
            fillcolor="rgba(255, 99, 71, 0.1)",
            # mode="lines",
            line_color="rgba(255,255,255,0)",
            name="±20% off-set",
            showlegend=True,
        ),
        row="all",
        col="all",
        exclude_empty_subplots=True,
    )

    return fig_phase


# -------------------------------
def _create_fig_heatmap(
    ens_name: str,
    phase: str,
    df_diff_stat_phase: pd.DataFrame,
    filter_largest: int,
    zmax: float,
) -> px.imshow:
    """Return heatmap plot."""

    zmin = -zmax

    df_pivot = df_diff_stat_phase.pivot(
        index="WELL", columns="DATE", values="DIFF_MEAN"
    )

    df_sorted = df_diff_stat_phase[["WELL", "DIFF_MEAN"]].copy()
    df_sorted["DIFF_MEAN"] = df_sorted.DIFF_MEAN.abs()
    df_sorted = df_sorted.groupby("WELL").max()
    df_sorted = df_sorted.sort_values(by=["DIFF_MEAN"], ascending=False)

    if filter_largest > 0:
        wells_largest_misfit = list(df_sorted.index)[:filter_largest]
        df_pivot = df_pivot[df_pivot.index.isin(wells_largest_misfit)]

    logging.debug(f"Dataframe pivot table, {ens_name} diff {phase} phase:\n{df_pivot}")

    fig_phase = px.imshow(
        df_pivot,
        color_continuous_scale=HEATMAP_COLOR,
        zmin=zmin,
        zmax=zmax,
    )
    fig_phase.update_layout(
        title_text=f"{ens_name} - {phase} cummulative misfit (mean) vs date",
        title_font_size=16,
    )
    fig_phase.update_traces(
        hoverongaps=False,
        hovertemplate="Date: %{x}"
        "<br>Well: %{y}"
        "<br>Difference: %{z:.3s}<extra></extra>",
    )

    return fig_phase


# -------------------------------
def average_arrow_annotation(mean_value: np.float64, yref: str = "y") -> Dict[str, Any]:
    decimals = 0
    if mean_value < 0.001:
        decimals = 5
    elif mean_value < 0.01:
        decimals = 4
    elif mean_value < 0.1:
        decimals = 3
    elif mean_value < 10:
        decimals = 2
    elif mean_value < 100:
        decimals = 1

    text = f"Total average: {mean_value:,.{decimals}f}"

    return {
        "x": 0.5,
        "y": mean_value,
        "xref": "paper",
        "yref": yref,
        "text": text,
        "showarrow": True,
        "align": "center",
        "arrowhead": 2,
        "arrowsize": 1,
        "arrowwidth": 1,
        "arrowcolor": "#636363",
        "ax": 20,
        "ay": -25,
    }



from typing import Optional, Sequence

import pandas as pd

# pylint: disable=line-too-long
from webviz_subsurface.plugins._simulation_time_series._views._subplot_view._utils.derived_vectors_accessor.derived_vectors_accessor import (
    DerivedVectorsAccessor,
)


class DerivedVectorsAccessorMock(DerivedVectorsAccessor):
    #####################################################e
    #
    # Interface methods raise NotImplementedError
    #
    #####################################################

    def has_provider_vectors(self) -> bool:
        raise NotImplementedError("Method not implemented for mock!")

    def has_per_interval_and_per_day_vectors(self) -> bool:
        raise NotImplementedError("Method not implemented for mock!")

    def has_vector_calculator_expressions(self) -> bool:
        raise NotImplementedError("Method not implemented for mock!")

    def get_provider_vectors_df(
        self, realizations: Optional[Sequence[int]] = None
    ) -> pd.DataFrame:
        raise NotImplementedError("Method not implemented for mock!")

    def create_per_interval_and_per_day_vectors_df(
        self, realizations: Optional[Sequence[int]] = None
    ) -> pd.DataFrame:
        raise NotImplementedError("Method not implemented for mock!")

    def create_calculated_vectors_df(
        self, realizations: Optional[Sequence[int]] = None
    ) -> pd.DataFrame:
        raise NotImplementedError("Method not implemented for mock!")


def test_create_valid_realizations_query() -> None:
    realizations = [2, 3, 4, 6, 7, 8]
    test_accessor = DerivedVectorsAccessorMock(accessor_realizations=realizations)

    # Filter query (get valid realization numbers)
    assert test_accessor.create_valid_realizations_query([1, 2, 3]) == [2, 3]
    assert test_accessor.create_valid_realizations_query([3, 4, 5, 6]) == [3, 4, 6]
    assert test_accessor.create_valid_realizations_query([3, 2, 7, 6]) == [3, 2, 7, 6]
    assert test_accessor.create_valid_realizations_query([1, 5]) == []

    # When all realizations are selected -> no filter, i.e. None
    assert test_accessor.create_valid_realizations_query(realizations) is None
    assert (
        test_accessor.create_valid_realizations_query([1, 2, 3, 4, 5, 6, 7, 8]) is None
    )



from .ensemble_fault_polygons_provider import (
    EnsembleFaultPolygonsProvider,
    EnsembleFaultPolygonsProviderFactory,
    FaultPolygonsAddress,
    FaultPolygonsServer,
    SimulatedFaultPolygonsAddress,
)
from .ensemble_summary_provider.ensemble_summary_provider import (
    EnsembleSummaryProvider,
    Frequency,
    VectorMetadata,
)
from .ensemble_summary_provider.ensemble_summary_provider_factory import (
    EnsembleSummaryProviderFactory,
)
from .ensemble_summary_provider.utils import get_matching_vector_names
from .ensemble_surface_provider import (
    EnsembleSurfaceProvider,
    EnsembleSurfaceProviderFactory,
    ObservedSurfaceAddress,
    QualifiedDiffSurfaceAddress,
    QualifiedSurfaceAddress,
    SimulatedSurfaceAddress,
    StatisticalSurfaceAddress,
    SurfaceAddress,
    SurfaceArrayMeta,
    SurfaceArrayServer,
    SurfaceImageMeta,
    SurfaceImageServer,
)
from .ensemble_table_provider import (
    ColumnMetadata,
    EnsembleTableProvider,
    EnsembleTableProviderFactory,
    EnsembleTableProviderImplArrow,
)
from .well_provider import WellProvider, WellProviderFactory, WellServer



from typing import List

import webviz_core_components as wcc
from dash import dcc, html
from dash.development.base_component import Component
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC


def list_to_options(values: List) -> List:
    return [{"value": val, "label": val} for val in values]


class ColorScale(SettingsGroupABC):
    class Ids(StrEnum):
        COLORMAP = "colormap"
        COLORRANGEMANUAL = "color-range-lock"
        COLORMIN = "color-min"
        COLORMAX = "color-max"

    def __init__(self) -> None:
        super().__init__("Color scale")
        self.colormaps = [
            "Physics",
            "Physics reverse",
            "Rainbow",
            "Rainbow reverse",
            "Porosity",
            "Permeability",
            "Seismic",
            "Time/Depth",
            "Stratigraphy",
            "Facies",
            "GasOilWater",
            "GasWater",
            "OilWater",
            "Accent",
            "Colors_set_1",
            "Colors_set_3",
        ]

    def layout(self) -> List[Component]:

        return [
            html.Div(
                children=[
                    wcc.Label("Color map"),
                    wcc.Dropdown(
                        id=self.register_component_unique_id(ColorScale.Ids.COLORMAP),
                        options=list_to_options(self.colormaps),
                        value=self.colormaps[0],
                        clearable=False,
                    ),
                    html.Div(
                        children=[
                            wcc.Checklist(
                                id=self.register_component_unique_id(
                                    ColorScale.Ids.COLORRANGEMANUAL
                                ),
                                options=[
                                    {
                                        "label": "Manual color range",
                                        "value": "manual_color",
                                    }
                                ],
                                value=[],
                                style={"marginTop": "10px"},
                                persistence=None,
                            ),
                            html.Div(
                                style={"display": "flex"},
                                children=[
                                    wcc.Label(
                                        style={"marginRight": "10px"}, children="Min: "
                                    ),
                                    dcc.Input(
                                        style={"maxWidth": "80px"},
                                        id=self.register_component_unique_id(
                                            ColorScale.Ids.COLORMIN
                                        ),
                                        type="number",
                                        minLength=1,
                                        disabled=True,
                                    ),
                                    wcc.Label(
                                        style={
                                            "marginLeft": "10px",
                                            "marginRight": "10px",
                                        },
                                        children="Max: ",
                                    ),
                                    dcc.Input(
                                        style={"maxWidth": "80px"},
                                        id=self.register_component_unique_id(
                                            ColorScale.Ids.COLORMAX
                                        ),
                                        type="number",
                                        minLength=1,
                                        disabled=True,
                                    ),
                                ],
                            ),
                        ],
                    ),
                ],
            ),
        ]



import datetime
from typing import List

import webviz_core_components as wcc
from dash.development.base_component import Component
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC


class WellOverviewSelections(SettingsGroupABC):
    class Ids(StrEnum):
        ENSEMBLES = "ensembles"
        RESPONSE = "response"
        PROD_FROM_DATE = "prod-from-date"
        PROD_UNTIL_DATE = "prod-until-date"

    def __init__(self, ensembles: List[str], dates: List[datetime.datetime]) -> None:
        super().__init__("Selections")

        self._ensembles = ensembles
        self._dates = dates

    def layout(self) -> List[Component]:
        return [
            wcc.Dropdown(
                label="Ensembles",
                id=self.register_component_unique_id(self.Ids.ENSEMBLES),
                options=[{"label": col, "value": col} for col in self._ensembles],
                value=self._ensembles,
                multi=True,
            ),
            wcc.Dropdown(
                label="Response",
                id=self.register_component_unique_id(self.Ids.RESPONSE),
                options=[
                    {"label": "Oil production", "value": "WOPT"},
                    {"label": "Gas production", "value": "WGPT"},
                    {"label": "Water production", "value": "WWPT"},
                ],
                value="WOPT",
                multi=False,
                clearable=False,
            ),
            wcc.Dropdown(
                label="Production From Date",
                id=self.register_component_unique_id(self.Ids.PROD_FROM_DATE),
                options=[
                    {
                        "label": dte.strftime("%Y-%m-%d"),
                        "value": dte.strftime("%Y-%m-%d"),
                    }
                    for dte in self._dates
                ],
                multi=False,
            ),
            wcc.Dropdown(
                label="Production Until Date",
                id=self.register_component_unique_id(self.Ids.PROD_UNTIL_DATE),
                options=[
                    {
                        "label": dte.strftime("%Y-%m-%d"),
                        "value": dte.strftime("%Y-%m-%d"),
                    }
                    for dte in self._dates
                ],
                multi=False,
            ),
        ]



from pathlib import Path

import pandas as pd
from webviz_config.common_cache import CACHE
from webviz_config.webviz_store import webvizstore

# The fmu.ensemble dependency ecl is only available for Linux,
# hence, ignore any import exception here to make
# it still possible to use the PvtPlugin on
# machines with other OSes.
#
# NOTE: Functions in this file cannot be used
#       on non-Linux OSes.
try:
    import fmu.ensemble
except ImportError:
    pass

from .fmu_input import load_ensemble_set


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def extract_mismatch(ens_paths: dict, observation_file: Path) -> pd.DataFrame:
    """Convert the fmu-ensemble mismatch dataframe into the the format
    suitable for the interactive history match visualization.
    """

    ens_data = load_ensemble_set(ens_paths)

    df_mismatch = fmu.ensemble.Observations(str(observation_file)).mismatch(ens_data)

    df_mismatch["NORMALISED_MISMATCH"] = df_mismatch["L2"] / (
        df_mismatch["MEASERROR"] ** 2
    )

    # Create a dataframe containing number of
    # observation points within each observation key:
    df_count = (
        df_mismatch.groupby(["OBSKEY", "REAL", "ENSEMBLE"])
        .size()
        .to_frame("COUNT")
        .reset_index()
        .drop_duplicates(["OBSKEY"], keep="first")
        .drop(columns=["REAL", "ENSEMBLE"])
    )

    # 1) Sum the normalised misfit (grouped by obskey, misfit sign
    #    realizaton and ensemble.
    # 2) Pivot the dataframe such that instead of two rows wrt. positive and
    #    negative misfit, we get two columns.
    # 3) Replace NaN values with 0 (NaN happens e.g. for the summed negative
    #    misfit if e.g. all misfit values are positive.
    # 4) Drop the column name 0 (webviz don't need summed misfit over all
    #    observation points with zero misfit :p)
    # 5) Merge in the COUNT column.
    # 6) Rename columns such that the columns from fmu.ensemble corresponds
    #    to those used in the webviz history match visualization.
    return (
        df_mismatch.groupby(["OBSKEY", "SIGN", "REAL", "ENSEMBLE"])
        .sum()[["NORMALISED_MISMATCH"]]
        .pivot_table(
            index=["OBSKEY", "REAL", "ENSEMBLE"],
            columns="SIGN",
            values="NORMALISED_MISMATCH",
        )
        .reset_index()
        .fillna(0)
        .drop(columns=[0], errors="ignore")
        .merge(df_count, on="OBSKEY", how="left")
        .rename(
            columns={
                "OBSKEY": "obs_group_name",
                "REAL": "realization",
                "ENSEMBLE": "ensemble_name",
                "COUNT": "number_data_points",
                1: "total_pos",
                -1: "total_neg",
            }
        )
    )
